from __future__ import annotations

import typing

from ctc import spec
from . import erc20_spec
from . import erc20_metadata


async def async_erc20_eth_call(
    function_name: str,
    token: spec.ERC20Reference,
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Any:
    """perform eth_call for an erc20"""

    from ctc import rpc

    address = await erc20_metadata.async_get_erc20_address(
        token, context=context
    )
    return await rpc.async_eth_call(
        to_address=address,
        function_abi=erc20_spec.erc20_function_abis[function_name],
        block_number=block,
        context=context,
        **rpc_kwargs,
    )


async def async_erc20s_eth_calls(
    function_name: str,
    tokens: typing.Iterable[spec.ERC20Reference],
    *,
    block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Sequence[typing.Any]:
    """perform eth_call for multiple erc20s"""

    import asyncio
    from ctc import rpc

    coroutines = [
        erc20_metadata.async_get_erc20_address(token, context=context)
        for token in tokens
    ]
    addresses = await asyncio.gather(*coroutines)
    return await rpc.async_batch_eth_call(
        to_addresses=addresses,
        function_abi=erc20_spec.erc20_function_abis[function_name],
        block_number=block,
        context=context,
        **rpc_kwargs,
    )


async def async_erc20_eth_call_by_block(
    function_name: str,
    token: spec.ERC20Reference,
    *,
    blocks: typing.Iterable[spec.BlockNumberReference],
    context: spec.Context = None,
    **rpc_kwargs: typing.Any,
) -> typing.Sequence[typing.Any]:
    """perform eth_call for an erc20 across multiple blocks"""

    from ctc import rpc

    address = await erc20_metadata.async_get_erc20_address(
        token, context=context
    )
    return await rpc.async_batch_eth_call(
        to_address=address,
        function_abi=erc20_spec.erc20_function_abis[function_name],
        block_numbers=blocks,
        context=context,
        **rpc_kwargs,
    )




"""
Maker multicall
- https://github.com/makerdao/multicall

Uniswap V3 multicall
- https://docs.uniswap.org/protocol/reference/periphery/base/Multicall
- https://github.com/Uniswap/v3-periphery/blob/main/contracts/base/Multicall.sol
"""

from __future__ import annotations

import asyncio
import typing

from ctc import config
from ctc import evm
from ctc import spec
from ctc import rpc
from . import call_utils
from . import multicall_spec


function_abis: typing.Mapping[str, spec.FunctionABI] = {
    'aggregate': {
        'inputs': [
            {
                'components': [
                    {'name': 'target', 'type': 'address'},
                    {'name': 'callData', 'type': 'bytes'},
                ],
                'name': 'calls',
                'type': 'tuple[]',
            }
        ],
        'name': 'aggregate',
        'outputs': [
            {'name': 'blockNumber', 'type': 'uint256'},
            {'name': 'returnData', 'type': 'bytes[]'},
        ],
        'stateMutability': 'nonpayable',
        'type': 'function',
    }
}


def get_multicall_address(
    *,
    context: spec.Context = None,
    version: str = 'maker',
) -> spec.Address:

    network = config.get_context_chain_id(context)
    network_name = evm.get_network_name(network)

    if version.lower() == 'Maker'.lower():
        multicall = {
            'ethereum': '0xeefba1e63905ef1d7acba5a8513c70307c1ce441',
            'kovan': '0x2cc8688c5f75e365aaeeb4ea8d6a480405a48d2a',
            'rinkeby': '0x42ad527de7d4e9d9d011ac45b31d8551f8fe9821',
            'gorli': '0x77dca2c955b15e9de4dbbcf1246b4b85b651e50e',
            'ropsten': '0x53c43764255c17bd724f74c4ef150724ac50a3ed',
            'xdai': '0xb5b692a88bdfc81ca69dcb1d924f59f0413a602a',
            'polygon': '0x11ce4b23bd875d7f5c6a31084f55fde1e9a87507',
            'mumbai': '0x08411add0b5aa8ee47563b146743c13b3556c9cc',
        }
        return multicall[network_name]
    elif version.lower() == 'Uniswap V3':
        # same across networks
        return '0x5ba1e12693dc8f9c48aad8770482f4739beed696'
    else:
        raise Exception('unknown version: ' + str(version))


async def async_multicall(
    calls: typing.Sequence[multicall_spec.Call],
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> typing.List[typing.Any]:

    # encode calls
    coroutines = [
        call_utils.async_encode_call(call, context=context) for call in calls
    ]
    encoded_calls = await asyncio.gather(*coroutines)

    # get multicall contract address
    multicall_address = get_multicall_address(context=context)

    # make call
    results = await rpc.async_eth_call(
        to_address=multicall_address,
        function_abi=function_abis['aggregate'],
        function_parameters=[encoded_calls],
        block_number=block,
        context=context,
    )
    block_number, encoded_outputs = results

    # decode outputs
    coroutines = [
        call_utils.async_decode_call_output(
            call=call,
            encoded_output=encoded_output,
            context=context,
        )
        for call, encoded_output in zip(calls, encoded_outputs)
    ]
    decoded_outputs = await asyncio.gather(*coroutines)

    return decoded_outputs


async def async_multicall_by_block(
    calls: typing.Sequence[multicall_spec.Call],
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    context: spec.Context = None,
) -> typing.Sequence[typing.Sequence[typing.Any]]:
    coroutines = [
        async_multicall(
            calls=calls,
            block=block,
            context=context,
        )
        for block in blocks
    ]
    results = asyncio.gather(*coroutines)
    return list(zip(*results))



from .block_coding import *
from .block_convert import *
from .block_crud import *
from .block_gas import *
from .block_hashes import *
from .block_normalize import *
from .block_prices import *
from .block_samples import *
from .block_summary import *
from .block_times import *



from __future__ import annotations

import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': decode_command,
        'help': """decode encoded EVM datatypes

for simple datatypes, no quotes required

for nested datatypes, enclose in quotes and quote contained addresses""",
        'args': [
            {'name': 'type', 'help': 'type of data to decode'},
            {'name': 'data', 'help': 'data to decode'},
        ],
        'examples': [
            'address 0x0000000000000000000000006b175474e89094c44da98b954eedeac495271d0f',
            '"(int64,int64,int64)" 0x000000000000000000000000000000000000000000000000000000000000000100000000000000000000000000000000000000000000000000000000000000020000000000000000000000000000000000000000000000000000000000000003',
        ],
    }


def decode_command(type: str, data: str) -> None:
    decoded = evm.abi_decode(evm.to_binary(data), type)
    print(decoded)



import toolsql

from ctc import db

import conftest


example_data = [
    {
        'address': '0x956f47f50a910163d8bf957cf5846d573e7f87ca',
        'symbol': 'FEI',
        'decimals': 18,
        'name': 'Fei USD',
    },
    {
        'address': '0xc7283b66eb1eb5fb86327f08e1b5816b0720212b',
        'symbol': 'TRIBE',
        'decimals': 18,
        'name': 'Tribe',
    },
]


async def test_erc20_metadata_crud():
    db_config = conftest.get_test_db_config()
    db_schema = db.get_prepared_schema(
        schema_name='erc20_metadata',
        context=dict(network='ethereum'),
    )
    toolsql.create_db(
        db_config=db_config,
        db_schema=db_schema,
        if_not_exists=True,
        confirm=True,
    )

    # insert data
    async with toolsql.async_connect(db_config) as conn:
        await db.async_upsert_erc20s_metadata(
            conn=conn,
            erc20s_metadata=example_data,
            context=dict(network=1),
        )

    # get data individually
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            actual_metadata = await db.async_select_erc20_metadata(
                conn=conn,
                address=datum['address'],
            )
            for key, target_value in datum.items():
                assert target_value == actual_metadata[key]

    # get data collectively
    all_addresses = [datum['address'] for datum in example_data]
    async with toolsql.async_connect(db_config) as conn:
        actual_metadatas = await db.async_select_erc20s_metadata(
            conn=conn,
            addresses=all_addresses,
        )
        sorted_example_data = sorted(example_data, key=lambda x: x['address'])
        sorted_actual_data = sorted(
            actual_metadatas, key=lambda x: x['address']
        )
        for target, actual in zip(sorted_example_data, sorted_actual_data):
            assert target == actual

    # delete entries one by one
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await db.async_delete_erc20_metadata(
                conn=conn,
                address=datum['address'],
                context=dict(network=1),
            )

    # ensure all entries deleted
    async with toolsql.async_connect(db_config) as conn:
        actual_metadatas = await db.async_select_erc20s_metadata(
            conn=conn,
            addresses=all_addresses,
        )
        assert all(item is None for item in actual_metadatas)

    # insert data again
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await db.async_upsert_erc20_metadata(
                conn=conn, context=dict(network=1), **datum
            )

    # delete entries all at once
    async with toolsql.async_connect(db_config) as conn:
        await db.async_delete_erc20s_metadata(
            conn=conn,
            addresses=all_addresses,
            context=dict(network=1),
        )

    # ensure all entries deleted
    async with toolsql.async_connect(db_config) as conn:
        actual_metadatas = await db.async_select_erc20s_metadata(
            conn=conn,
            addresses=all_addresses,
        )
        assert all(item is None for item in actual_metadatas)




from __future__ import annotations

import typing

import aiohttp

if typing.TYPE_CHECKING:
    from typing_extensions import Literal
    from typing_extensions import TypedDict

    class TVLTimeseries(TypedDict):
        timestamp: typing.Sequence[int]
        tvl: typing.Sequence[int | float]

    class YieldTimeseries(TypedDict):
        timestamp: typing.Sequence[int]
        tvl: typing.Sequence[int | float]
        apy: typing.Sequence[int | float]


tvl_url_root = 'https://api.llama.fi'
yields_url_root = 'https://yields.llama.fi'

url_templates = {
    'historical_defi_tvl': '/charts',
    'protocols_tvls': '/protocols',
    'historical_protocol_tvl': '/protocol/{protocol}',
    'chains_tvls': '/chains',
    'historical_chain_tvl': '/charts/{chain}',
    'pools': '/pools',
    'pool_yield': '/chart/{pool}',
}


async def async_request_llama_data(
    datatype: str, **parameters: typing.Any
) -> typing.Any:

    if datatype in ['pools', 'pool_yield']:
        root = yields_url_root
    else:
        root = tvl_url_root

    # create url
    if parameters is None:
        parameters = {}
    url = root + url_templates[datatype].format(**parameters)

    # make request
    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            return await response.json()


async def async_get_historical_defi_tvl() -> TVLTimeseries:
    data = await async_request_llama_data('historical_defi_tvl')

    dates = []
    tvls = []
    for datum in data:
        dates.append(int(datum['date']))
        tvls.append(datum['totalLiquidityUSD'])

    return {'timestamp': dates, 'tvl': tvls}


async def async_get_protocols_tvls(
    *,
    n: int | None = None,
    category: str | None = None,
    chain: str | None = None,
) -> typing.Sequence[typing.Any]:
    data = await async_request_llama_data('protocols_tvls')

    filtered = []
    for datum in data:

        if category is not None and datum['category'] != category:
            continue

        if chain is not None and chain not in datum['chains']:
            continue

        filtered.append(datum)

        if n is not None and len(filtered) >= n:
            break

    return filtered


async def async_get_historical_protocol_tvl(protocol: str) -> typing.Any:

    data = await async_request_llama_data(
        'historical_protocol_tvl', protocol=protocol
    )

    return data


async def async_get_chains_tvls() -> typing.Sequence[typing.Any]:
    data: typing.Sequence[typing.Any] = await async_request_llama_data('chains_tvls')

    return data


async def async_get_historical_chain_tvl(chain: str) -> TVLTimeseries:

    data = await async_request_llama_data('historical_chain_tvl', chain=chain)

    timestamp = []
    tvl = []

    for datum in data:
        timestamp.append(int(datum['date']))
        tvl.append(datum['totalLiquidityUSD'])

    return {'timestamp': timestamp, 'tvl': tvl}


async def async_get_llama_pools(
    *,
    chain: str | None = None,
    project: str | None = None,
    stablecoin: bool | None = None,
    exposure: Literal['single', 'multi'] | None = None,
    min_apy: typing.SupportsFloat | None = None,
    max_apy: typing.SupportsFloat | None = None,
    min_tvl: typing.SupportsFloat | None = None,
    max_tvl: typing.SupportsFloat | None = None,
) -> typing.Sequence[typing.Any]:

    data = await async_request_llama_data('pools')

    if min_apy is not None:
        min_apy = float(min_apy)
    if max_apy is not None:
        max_apy = float(max_apy)
    if min_tvl is not None:
        min_tvl = float(min_tvl)
    if max_tvl is not None:
        max_tvl = float(max_tvl)

    pools = []
    for pool in data['data']:
        if chain is not None and pool['chain'] != chain:
            continue
        if project is not None and pool['project'] != project:
            continue
        if stablecoin is not None and pool['stablecoin'] != stablecoin:
            continue
        if exposure is not None and pool['exposure'] != exposure:
            continue
        if min_apy is not None and pool['apy'] < min_apy:
            continue
        if max_apy is not None and pool['apy'] > max_apy:
            continue
        if min_tvl is not None and pool['tvlUsd'] < min_tvl:
            continue
        if max_tvl is not None and pool['tvlUsd'] > max_tvl:
            continue

        pools.append(pool)

    return pools


async def async_get_pool_yield(pool: str) -> YieldTimeseries:

    import tooltime

    response = await async_request_llama_data('pool_yield', pool=pool)

    timestamp = []
    tvl = []
    apy = []
    for datum in response['data']:
        timestamp.append(tooltime.timestamp_to_seconds(datum['timestamp']))
        tvl.append(datum['tvlUsd'])
        apy.append(datum['apy'])

    return {
        'timestamp': timestamp,
        'tvl': tvl,
        'apy': apy,
    }



from __future__ import annotations

import typing

from ctc import spec

from .. import rpc_constructors
from .. import rpc_digestors
from .. import rpc_request


def sync_eth_new_filter(
    *,
    address: spec.Address | None = None,
    topics: typing.Sequence[spec.BinaryData | None] | None = None,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
    decode_response: bool = False,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_new_filter(
        address=address,
        topics=topics,
        start_block=start_block,
        end_block=end_block,
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_new_filter(
        response=response,
        decode_response=decode_response,
    )


def sync_eth_new_block_filter(
    *, context: spec.Context = None, decode_response: bool = False
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_new_block_filter()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_new_block_filter(
        response=response,
        decode_response=decode_response,
    )


def sync_eth_new_pending_transaction_filter(
    *, context: spec.Context = None, decode_response: bool = False
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_new_pending_transaction_filter()
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_new_pending_transaction_filter(
        response=response,
        decode_response=decode_response,
    )


def sync_eth_uninstall_filter(
    filter_id: spec.GenericBinaryData,
    *,
    context: spec.Context = None,
    decode_response: bool = False,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_uninstall_filter(
        filter_id=filter_id
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_uninstall_filter(
        response=response,
        decode_response=decode_response,
    )


def sync_eth_get_filter_changes(
    filter_id: spec.GenericBinaryData,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
    include_removed: bool = False,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_filter_changes(
        filter_id=filter_id
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_filter_changes(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
        include_removed=include_removed,
    )


def sync_eth_get_filter_logs(
    filter_id: spec.GenericBinaryData,
    *,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
    include_removed: bool = False,
) -> spec.RpcSingularResponse:
    request = rpc_constructors.construct_eth_get_filter_logs(
        filter_id=filter_id
    )
    response = rpc_request.sync_send(request, context=context)
    return rpc_digestors.digest_eth_get_filter_logs(
        response=response,
        decode_response=decode_response,
        snake_case_response=snake_case_response,
        include_removed=include_removed,
    )


def sync_eth_get_logs(
    *,
    address: spec.BinaryData | None = None,
    topics: typing.Sequence[spec.BinaryData | None] | None = None,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    block_hash: spec.BinaryData | None = None,
    context: spec.Context = None,
    decode_response: bool = True,
    snake_case_response: bool = True,
    include_removed: bool = False,
) -> spec.RpcSingularResponse:

    request = rpc_constructors.construct_eth_get_logs(
        address=address,
        topics=topics,
        start_block=start_block,
        end_block=end_block,
        block_hash=block_hash,
    )

    import sys

    if (sys.version_info.major, sys.version_info.minor) >= (3, 8):
        from ctc.rpc.rpc_decoders import log_decoder

        raw_response = rpc_request.sync_send(
            request,
            context=context,
            raw_output=True,
        )
        return log_decoder.decode_logs(
            raw_response,
            include_removed=include_removed,
        )

    else:
        response = rpc_request.sync_send(request, context=context)
        logs = rpc_digestors.digest_eth_get_logs(
            response=response,
            decode_response=decode_response,
            snake_case_response=snake_case_response,
            include_removed=include_removed,
        )
        return [
            (
                log['block_number'],
                log['transaction_index'],
                log['log_index'],
                log['transaction_hash'],
                log['address'],
                tuple(log['topics']),
                log['data'],
                log['block_hash'],
            )
            for log in logs
        ]




from __future__ import annotations

import typing

from ctc import spec
from . import twap_spec


async def async_get_chainlink_data(
    data_source: twap_spec.DataSource,
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    context: spec.Context = None,
) -> spec.DataFrame:

    from ctc.protocols import chainlink_utils

    feed = data_source.get('feed')
    composite_feed = data_source.get('composite_feed')
    invert = data_source.get('invert')
    if invert is None:
        invert = False
    normalize = data_source.get('normalize')
    if normalize is None:
        normalize = False

    if feed is not None:
        return await chainlink_utils.async_get_feed_data(
            feed=feed,
            invert=invert,
            normalize=normalize,
            start_block=start_block,
            end_block=end_block,
            interpolate=True,
            fields='answer',
            context=context,
        )
    elif composite_feed is not None:
        return await chainlink_utils.async_get_composite_feed_data(
            composite_feed=composite_feed,
            invert=invert,
            start_block=start_block,
            end_block=end_block,
            context=context,
        )
    else:
        raise Exception('must specify feed or composite_feed')


# async def async_get_uniswap_v2_data(
#     data_source: twap_spec.DataSource,
#     start_block: typing.Optional[spec.BlockNumberReference] = None,
#     end_block: typing.Optional[spec.BlockNumberReference] = None,
#     context: spec.Context = None,
# ) -> spec.Series:

#     from ctc.protocols import uniswap_v2_utils

#     feed = data_source.get('feed')
#     composite_feed = data_source.get('composite_feed')
#     invert = data_source.get('invert')
#     normalize = data_source.get('normalize')

#     if feed is not None:
#         return await uniswap_v2_utils.async_get_feed_data(
#             feed=feed,
#             invert=invert,
#             normalize=normalize,
#             start_block=start_block,
#             end_block=end_block,
#             interpolate=True,
#             context=context,
#         )
#     elif composite_feed is not None:
#         return await uniswap_v2_utils.async_get_composite_feed_data(
#             composite_feed=composite_feed,
#             invert=invert,
#             normalize=normalize,
#             start_block=start_block,
#             end_block=end_block,
#             context=context,
#         )
#     else:
#         raise Exception('must specify feed or composite_feed')



from __future__ import annotations

import typing
import time

from ctc import spec
from .. import binary_utils
from .. import transaction_utils
from . import block_crud


async def async_print_block_summary(
    block: spec.DBBlock | spec.BlockNumberReference,
    *,
    context: spec.Context = None,
) -> None:
    """print summary of block"""

    import toolstr
    import tooltime

    from ctc import cli

    if not isinstance(block, dict):
        block = await block_crud.async_get_block(
            block=block, context=context
        )

    block_transactions = await transaction_utils.async_get_block_transactions(
        block['number'], context=context
    )
    percentiles = [
        0,
        5,
        50,
        95,
        100,
    ]

    gas_prices = []
    for transaction in block_transactions:
        gas_prices.append(transaction['gas_price'] / 1e9)
    import numpy as np

    gas_percentiles: typing.Sequence[typing.Any] | spec.NumpyArray
    if len(gas_prices) > 0:
        gas_percentiles = np.percentile(
            gas_prices,
            percentiles,
        )
    else:
        gas_percentiles = [None] * len(percentiles)

    title = 'Block ' + str(block['number'])
    styles = cli.get_cli_styles()
    toolstr.print_text_box(title, style=styles['title'])
    cli.print_bullet(key='timestamp', value=block['timestamp'])
    cli.print_bullet(
        key='time', value=tooltime.timestamp_to_iso(block['timestamp'])
    )
    cli.print_bullet(
        key='age',
        value=tooltime.timelength_to_phrase(
            round(time.time()) - block['timestamp']
        ),
    )
    cli.print_bullet(key='block_hash', value=block['hash'])
    cli.print_bullet(key='n_transactions', value=len(block_transactions))
    cli.print_bullet(
        key='gas used',
        value=(
            toolstr.format(block['gas_used'])
            + ' / '
            + toolstr.format(block['gas_limit'])
        ),
    )

    if block_transactions is not None:
        percentile_label = (
            '('
            + ', '.join([str(percentile) + '%' for percentile in percentiles])
            + ')'
        )
        gas_percentiles_str = (
            '('
            + ', '.join(
                [toolstr.format(percentile) for percentile in gas_percentiles]
            )
            + ')'
        )
        cli.print_bullet(
            key='gas prices:',
            value=percentile_label + '=' + gas_percentiles_str,
        )

    message = block['extra_data']
    try:
        message = binary_utils.to_binary(message).decode()
    except Exception:
        if len(message) > 80:
            message = message[:77] + '...'
        else:
            message = message
    cli.print_bullet(key='message', value=message)




import os
import tempfile

import pytest
import toolsql

from ctc import db
from ctc.protocols.chainlink_utils import chainlink_db
from ctc.protocols import fourbyte_utils
from ctc.protocols.coingecko_utils import coingecko_db

import conftest


schema_datas = [
    {
        'schema_name': 'block_timestamps',
        'selector': db.schemas.block_timestamps.async_select_block_timestamp,
        'queryer': db.schemas.block_timestamps.async_query_block_timestamp,
        'query': {'block_number': 100},
        'plural_selector': db.schemas.block_timestamps.async_select_block_timestamps,
        'plural_queryer': db.schemas.block_timestamps.async_query_block_timestamp,
        'plural_query': {'block_numbers': [100, 200, 300]},
    },
    {
        'schema_name': 'block_timestamps',
        'selector': db.schemas.block_timestamps.async_select_max_block_timestamp,
        'queryer': db.schemas.block_timestamps.async_query_max_block_timestamp,
        'query': {},
    },
    {
        'schema_name': 'block_timestamps',
        'selector': db.schemas.blocks.blocks_statements.async_select_max_block_number,
        # 'queryer': db.schemas.blocks.blocks_statements.async_query_max_block_number,
        'query': {},
    },
    {
        'schema_name': 'block_timestamps',
        'selector': db.schemas.blocks.blocks_statements.async_select_max_block_timestamp,
        # 'queryer': db.schemas.blocks.blocks_statements.async_query_max_block_timestamp,
        'query': {},
    },
    {
        'schema_name': 'blocks',
        'selector': db.schemas.blocks.async_select_block,
        'queryer': db.schemas.blocks.async_query_block,
        'query': {'block_number': 100},
        'plural_selector': db.schemas.blocks.async_select_blocks,
        'plural_queryer': db.schemas.blocks.async_query_blocks,
        'plural_query': {'block_numbers': [100, 200, 300]},
    },
    {
        'schema_name': 'contract_abis',
        'selector': db.async_select_contract_abi,
        'queryer': db.async_query_contract_abi,
        'query': {'address': '0x956f47f50a910163d8bf957cf5846d573e7f87ca'},
        'plural_selector': db.async_select_contract_abis,
        'plural_queryer': db.async_query_contract_abis,
        'plural_query': {},
    },
    {
        'schema_name': 'contract_creation_blocks',
        'selector': db.async_select_contract_creation_block,
        'queryer': db.async_query_contract_creation_block,
        'query': {'address': '0x956f47f50a910163d8bf957cf5846d573e7f87ca'},
        'plural_selector': db.async_select_contract_creation_blocks,
        'plural_queryer': db.async_query_contract_creation_blocks,
        'plural_query': {},
    },
    {
        'schema_name': 'erc20_metadata',
        'selector': db.async_select_erc20_metadata,
        'queryer': db.async_query_erc20_metadata,
        'query': {'address': '0x956f47f50a910163d8bf957cf5846d573e7f87ca'},
        'plural_selector': db.async_select_erc20s_metadata,
        'plural_queryer': db.async_query_erc20s_metadata,
        'plural_query': {
            'addresses': ['0x956f47f50a910163d8bf957cf5846d573e7f87ca'],
        },
    },
    {
        'schema_name': 'dex_pools',
        'selector': db.async_select_dex_pool,
        'queryer': db.async_query_dex_pool,
        'query': {'address': '0x9928e4046d7c6513326ccea028cd3e7a91c7590a'},
        'plural_selector': db.async_select_dex_pools,
        'plural_queryer': db.async_query_dex_pools,
        'plural_query': {
            'factory': '0x5c69bee701ef814a2b6a3edd4b1652cb9cc5aa6f',
        },
    },
    {
        'schema_name': 'chainlink',
        'selector': chainlink_db.async_select_feed,
        'queryer': chainlink_db.async_query_feed,
        'query': {'address': '0x31e0a88fecb6ec0a411dbe0e9e76391498296ee9'},
        'plural_selector': chainlink_db.async_select_feeds,
        'plural_queryer': chainlink_db.async_query_feeds,
        'plural_query': {
            'addresses': ['0x31e0a88fecb6ec0a411dbe0e9e76391498296ee9'],
        },
    },
    {
        'schema_name': '4byte',
        # 'selector': fourbyte_utils.async_select_function_entries,
        # 'queryer': fourbyte_utils.async_query_function_entries,
        # 'query': {'address': '0x31e0a88fecb6ec0a411dbe0e9e76391498296ee9'},
        'plural_selector': fourbyte_utils.async_select_function_signatures,
        'plural_queryer': fourbyte_utils.async_query_function_signatures,
        'plural_query': {
            'text_signature': 'transfer(address,uint256)',
        },
    },
    # {
    #     'schema_name': 'coingecko',
    #     'selector': coingecko_db.async_select_token,
    #     'queryer': coingecko_db.async_query_token,
    #     'query': {'id': '1244-s-avers'},
    #     'plural_selector': coingecko_db.async_select_tokens,
    #     'plural_queryer': coingecko_db.async_query_tokens,
    #     'plural_query': {
    #         'symbol_query': 'eth',
    #     },
    # },
    {
        'schema_name': 'block_gas',
        'selector': db.async_select_median_block_gas_fee,
        'queryer': db.async_query_median_block_gas_fee,
        'query': {'block_number': 14000000},
        'plural_selector': db.async_select_median_blocks_gas_fees,
        'plural_queryer': db.async_query_median_blocks_gas_fees,
        'plural_query': {'block_numbers': [14000000, 14000001, 14000002]},
    },
    {
        'schema_name': 'events',
        'plural_selector': db.async_select_event_queries,
        'plural_queryer': db.async_query_event_queries,
        'plural_query': {
            'query_type': 3,
            'contract_address': '0x6b175474e89094c44da98b954eedeac495271d0f',
            'start_block': 14000000,
            'end_block': 14000000,
        },
    },
    {
        'schema_name': 'transactions',
        'selector': db.async_select_transaction,
        'queryer': db.async_query_transaction,
        'query': {
            'hash': '0xaef402056d26796c49a7cdc0be9fa09f193e48458a1829dd788c38a8ae143683'
        },
        'plural_selector': db.async_select_transactions,
        'plural_queryer': db.async_query_transactions,
        'plural_query': {
            'hashes': [
                '0xaef402056d26796c49a7cdc0be9fa09f193e48458a1829dd788c38a8ae143683',
                '0xb451a6aa6a5a0f30b62ea616f581b8f4c9a2e52091a6787591c5941be9995f75',
                '0x3ccc6a654fd68a2a2303c9ef2a54512890c96376016a50c83f227ef705d27a05',
            ]
        },
    },
]

non_network_schemas = (
    db.get_generic_schema_names() + db.get_admin_schema_names()
)


def test_all_evm_schemas_tested_for_problems():
    tested = {schema_datum['schema_name'] for schema_datum in schema_datas}
    tested |= {'coingecko'}
    assert tested == set(
        db.get_network_schema_names() + db.get_generic_schema_names()
    ), 'not all EVM schema types being tested for common db problems'


@pytest.mark.parametrize('schema_data', schema_datas)
async def test_select_when_db_folder_does_not_exist(schema_data):

    dne_dir = os.path.join(tempfile.mkdtemp(), 'does', 'not', 'exist')
    assert not os.path.isdir(dne_dir)
    dne_path = os.path.join(dne_dir, 'ctc.db')

    db_config = {'dbms': 'sqlite', 'path': dne_path}

    if schema_data.get('queryer') is not None:
        result = await schema_data['queryer'](
            context=dict(network=1),
            db_config=db_config,
            **schema_data['query'],
        )

        assert result is None


@pytest.mark.parametrize('schema_data', schema_datas)
async def test_select_when_db_file_does_not_exist(schema_data):
    db_config = conftest.get_test_db_config()
    if schema_data.get('queryer') is not None:
        result = await schema_data['queryer'](
            context=dict(network=1), db_config=db_config, **schema_data['query']
        )

        assert result is None


@pytest.mark.parametrize('schema_data', schema_datas)
async def test_select_when_schema_not_initialized(schema_data):

    db_config = conftest.get_test_db_config()
    with toolsql.connect(db_config) as conn:
        db.initialize_schema_versions(conn=conn)

    if schema_data['schema_name'] not in non_network_schemas:
        network_kwargs = {'context': {'network': 1}}
    else:
        network_kwargs = {}

    if schema_data.get('selector') is not None:
        with pytest.raises(toolsql.TableDoesNotExist):
            async with toolsql.async_connect(db_config) as conn:
                await schema_data['selector'](
                    conn=conn, **schema_data['query'], **network_kwargs
                )

    if 'plural_selector' in schema_data:
        with pytest.raises(toolsql.TableDoesNotExist):
            async with toolsql.async_connect(db_config) as conn:
                await schema_data['plural_selector'](
                    conn=conn,
                    **schema_data['plural_query'],
                    **network_kwargs,
                )


@pytest.mark.parametrize('schema_data', schema_datas)
async def test_query_when_schema_not_initialized(schema_data):
    db_config = conftest.get_test_db_config()
    with toolsql.connect(db_config) as conn:
        db.initialize_schema_versions(conn=conn)

    if schema_data['schema_name'] not in non_network_schemas:
        network_kwargs = {'context': {'network': 1}}
    else:
        network_kwargs = {'context': None}

    if schema_data.get('queryer') is not None:
        result = await schema_data['queryer'](
            db_config=db_config,
            **schema_data['query'],
            **network_kwargs,
        )

        assert result is None


# @pytest.mark.parametrize('schema_data', schema_datas)
# async def test_select_when_row_does_not_exist(schema_data):
#     db_config = conftest.get_test_db_config()
#     with toolsql.connect(db_config) as conn:
#         if schema_data['schema_name'] not in non_network_schemas:
#             network_kwargs = {'context': {'network': 1}}
#         else:
#             network_kwargs = {'context': {'network': None}}

#         db.initialize_schema(
#             schema_name=schema_data['schema_name'], conn=conn, **network_kwargs
#         )

#     if schema_data['schema_name'] not in non_network_schemas:
#         network_kwargs = {'context': {'network': 1}}
#     else:
#         network_kwargs = {}

#     if schema_data.get('selector') is not None:
#         async with toolsql.async_connect(db_config) as conn:
#             result = await schema_data['selector'](
#                 conn=conn, **schema_data['query'], **network_kwargs
#             )

#         assert result is None

#     if 'plural_selector' in schema_data:
#         async with toolsql.async_connect(db_config) as conn:
#             result = await schema_data['plural_selector'](
#                 conn=conn,
#                 **schema_data['plural_query'],
#                 **network_kwargs,
#             )

#         assert result is None or all(item is None for item in result)


@pytest.mark.parametrize('schema_data', schema_datas)
async def test_query_when_row_does_not_exist(schema_data):
    db_config = conftest.get_test_db_config()
    with toolsql.connect(db_config) as conn:
        if schema_data['schema_name'] not in non_network_schemas:
            network_kwargs = {'context': {'network': 1}}
        else:
            network_kwargs = {'context': {'network': None}}

        db.initialize_schema(
            schema_name=schema_data['schema_name'],
            conn=conn,
            **network_kwargs,
        )

    if schema_data.get('queryer') is not None:
        if schema_data['schema_name'] not in non_network_schemas:
            network_kwargs = {'context': {'network': 1}}
        else:
            network_kwargs = {'context': None}

        result = await schema_data['queryer'](
            db_config=db_config,
            **schema_data['query'],
            **network_kwargs,
        )

        assert result is None




from __future__ import annotations

import typing


class BadSearchDomain(Exception):
    pass


async def async_bisect(
    async_f: typing.Callable[
        ...,
        typing.Coroutine[typing.Any, typing.Any, float],
    ],
    *,
    a: int | float,
    b: int | float,
    input_tol: float | None = None,
    output_tol: float | None = None,
    max_iterations: int | None = None,
    f_args: typing.Sequence[typing.Any] | None = None,
    f_kwargs: typing.Mapping[str, typing.Any] | None = None,
    verbose: bool = False,
) -> float:
    import asyncio

    # ensure that bounds are valid
    if a > b:
        raise BadSearchDomain(
            'a must be less than b (' + str(a) + ' >= ' + str(b) + ')'
        )

    # set defaults
    if f_args is None:
        f_args = []
    if f_kwargs is None:
        f_kwargs = {}

    # compute values on bounds
    f_a, f_b = await asyncio.gather(
        async_f(a, *f_args, **f_kwargs),
        async_f(b, *f_args, **f_kwargs),
    )

    if verbose:
        print('probing min: f(' + str(a) + ') = ' + str(f_a))
        print('probing max: f(' + str(b) + ') = ' + str(f_b))

    # check if solution on boundary
    if f_a == 0:
        return a
    if f_b == 0:
        return b

    # check that function has opposite signs at bounds
    if (f_a < 0 and f_b < 0) or (f_a > 0 and f_b > 0):
        raise BadSearchDomain(
            'function must have opposite signs at bounds, have:\n    f(a) = '
            + str(f_a)
            + '\n    f(b) = '
            + str(f_b)
        )

    iteration = 0
    while True:

        # compute c at midpoint
        c = (a + b) / 2
        f_c = await async_f(c, *f_args, **f_kwargs)

        if verbose:
            print('probing:   f(' + str(c) + ') = ' + str(f_c))

        # check if solution found
        if f_c == 0:
            break
        if input_tol is not None:
            input_tol_satisfied = (b - a) / 2 < input_tol
        else:
            input_tol_satisfied = True
        if output_tol is not None:
            output_tol_satisfied = abs(f_c) < output_tol
        else:
            output_tol_satisfied = True
        if input_tol_satisfied and output_tol_satisfied:
            break

        # check whether to update a or b
        if (f_c > 0) == (f_a > 0):
            a = c
            f_a = f_c
        else:
            b = c
            f_b = f_c

        if verbose:
            print('new range:', [a, b])

        iteration += 1
        if max_iterations is not None and iteration >= max_iterations:
            raise Exception('solution could not be found')

    if verbose:
        summary = '\n'
        summary += 'found solution ' + str(c)
        summary += '\n- iterations = ' + str(iteration + 1)
        if input_tol is not None:
            summary += (
                '\n- input_precision = '
                + str((b - a) / 2)
                + ' < '
                + str(input_tol)
                + ' = input_tol'
            )
        if output_tol is not None:
            summary += (
                '\n- output_precision = '
                + str(abs(f_c))
                + ' < '
                + str(input_tol)
                + ' = output_tol'
            )
        print(summary)

    return c



from __future__ import annotations

import asyncio
import json
import time
import typing

from ctc import config
from ctc import evm
from ctc import spec

from . import url_crud

if typing.TYPE_CHECKING:
    from typing_extensions import TypedDict

    class EtherscanRatelimit(TypedDict):
        requests_per_second: int | float
        last_request_time: int | float
        lock: asyncio.Lock | None
        recent_results: typing.MutableMapping[spec.Address, spec.ContractABI]


_etherscan_ratelimit: EtherscanRatelimit = {
    'requests_per_second': 0.2,
    'last_request_time': 0,
    'lock': None,
    'recent_results': {},
}


def set_etherscan_ratelimit(requests_per_second: int | float) -> None:
    _etherscan_ratelimit['requests_per_second'] = requests_per_second


async def async_get_contract_abi(
    contract_address: spec.Address,
    *,
    context: spec.Context = None,
    verbose: bool = True,
) -> spec.ContractABI:
    """fetch contract abi using etherscan"""

    import aiohttp

    network = config.get_context_chain_id(context)

    # process inputs
    if not evm.is_address_str(contract_address):
        raise Exception('not a valid address: ' + str(contract_address))

    # create lock
    lock = _etherscan_ratelimit['lock']
    if lock is None:
        lock = asyncio.Lock()
        _etherscan_ratelimit['lock'] = lock

    # acquire lock
    async with lock:

        if contract_address in _etherscan_ratelimit['recent_results']:
            return _etherscan_ratelimit['recent_results'][contract_address]

        # ratelimit
        now = time.time()
        time_since_last = now - _etherscan_ratelimit['last_request_time']
        seconds_per_request = 1 / _etherscan_ratelimit['requests_per_second']
        time_to_sleep = seconds_per_request - time_since_last
        if time_to_sleep > 0:
            if verbose:
                print(
                    'etherscan ratelimit hit, sleeping for '
                    + str(time_to_sleep)
                    + ' seconds'
                )
            await asyncio.sleep(time_to_sleep)

        if verbose:
            network_name = evm.get_network_name(network)
            print(
                'fetching ' + str(network_name) + ' abi from etherscan:',
                contract_address,
            )

        # create url
        abi_endpoint = url_crud.create_abi_url(
            address=contract_address,
            network=network,
        )

        # make request
        async with aiohttp.ClientSession() as session:
            async with session.get(abi_endpoint) as response:
                content = await response.text()
        _etherscan_ratelimit['last_request_time'] = time.time()

        # process request
        if content == 'Contract source code not verified':
            raise spec.AbiNotFoundException(
                'contract not verified on etherscan ' + str(contract_address)
            )
        abi = json.loads(content)
        if isinstance(abi, dict) and abi.get('status') == '0':
            raise Exception(
                'could not obtain contract abi from etherscan for '
                + str(contract_address)
            )

        _etherscan_ratelimit['recent_results'][contract_address] = abi

    if typing.TYPE_CHECKING:
        return typing.cast(spec.ContractABI, abi)
    else:
        return abi



"""functions related to file partitions

TODO:
- skipping files that already exist
- add partition column to name
"""

from __future__ import annotations

import glob
import os
import typing
import tempfile

import polars as pl
import toolstr

from ctc.toolbox import pl_utils
from ctc.toolbox import range_utils


def create_partition_bounds(
    n_partitions: int,
    n_bound_nibbles: int | None,
) -> typing.Sequence[int]:
    if n_bound_nibbles is None:
        if n_partitions <= 16:
            n_bound_nibbles = 1
        elif n_partitions <= 256:
            n_bound_nibbles = 2
        elif n_partitions <= 65536:
            n_bound_nibbles = 4
        elif n_partitions <= 16777216:
            n_bound_nibbles = 6
        elif n_partitions <= 4294967296:
            n_bound_nibbles = 8
        else:
            raise Exception('must specify number of nibbles')

    max_bound = 16**n_bound_nibbles
    interval_size = int(max_bound / n_partitions)
    bounds = list(range(0, max_bound + interval_size, interval_size))
    bounds = [value - 1 for value in bounds]
    return bounds


def repartition_files(
    *,
    method: typing.Literal['simple', 'chunked'],
    target: str | typing.Sequence[str],
    column: str,
    output_dir: str,
    input_template: str,
    output_template: str,
    template_aggregations: typing.Mapping[str, str],
    n_partitions: int,
    n_bound_nibbles: int,
    chunk_size: int | None = None,
    temporary_dir: str | None = None,
    delete_temporary: bool = True,
    n_row_groups_per_partition: int | None = None,
    row_group_size: int | None = None,
    verbose: bool | int = True,
    skip_existing: bool = False,
    columns_in_partition_names: bool = True,
) -> None:
    """
    Reindexing methods
    - simple
        - allowed when dataset fits in memory
        - uses 1 pass over the data for reading and writing
        - steps:
            - load all dataset files
            - sort data
            - partition data into files
    - chunked
        - required when dataset does not fit in memory
            - each chunk should fit in memory
        - uses 2 passes over the data for reading and writing
        - raw_files --> temporary_files --> final_files
        - for chunks of N dataset files:
            - load N dataset files
            - sort data
            - partition data into temporary files
        - for each partition of temporary files
            - load files
            - combine into single file
    """

    import time

    start = time.time()

    # get file chunks and temporary directory
    if isinstance(target, str):
        all_files = glob.glob(target)
        all_files = sorted(all_files)
    elif isinstance(target, list):
        all_files = target
    else:
        raise Exception('invalid target format')

    # create partition bounds and labels
    bounds = create_partition_bounds(
        n_partitions=n_partitions,
        n_bound_nibbles=n_bound_nibbles,
    )
    labels = [
        '0x'
        + hex(start + 1)[2:].rjust(n_bound_nibbles, '0')
        + '_to_0x'
        + hex(end)[2:].rjust(n_bound_nibbles, '0')
        for start, end in zip(bounds[:-1], bounds[1:])
    ]

    # create partitions
    if method == 'simple':
        _files_to_partitions(
            paths=all_files,
            bounds=bounds,
            labels=labels,
            column=column,
            output_dir=output_dir,
            input_template=input_template,
            output_template=output_template,
            skip_existing=skip_existing,
            columns_in_partition_names=columns_in_partition_names,
        )
    elif method == 'chunked':
        _repartition_files_chunked(
            all_files=all_files,
            bounds=bounds,
            labels=labels,
            column=column,
            input_template=input_template,
            output_template=output_template,
            template_aggregations=template_aggregations,
            output_dir=output_dir,
            n_bound_nibbles=n_bound_nibbles,
            n_partitions=n_partitions,
            chunk_size=chunk_size,
            temporary_dir=temporary_dir,
            delete_temporary=delete_temporary,
            n_row_groups_per_partition=n_row_groups_per_partition,
            row_group_size=row_group_size,
            verbose=verbose,
            skip_existing=skip_existing,
            columns_in_partition_names=columns_in_partition_names,
        )
    else:
        raise Exception('invalid method: ' + str(method))

    if verbose:
        end = time.time()
        print('took', '%.2f' % (end - start), 'seconds')


def _repartition_files_chunked(
    *,
    all_files: typing.Sequence[str],
    bounds: typing.Sequence[int],
    labels: typing.Sequence[str],
    column: str,
    output_dir: str,
    n_partitions: int,
    n_bound_nibbles: int | None,
    chunk_size: int | None,
    temporary_dir: str | None,
    delete_temporary: bool,
    template_aggregations: typing.Mapping[str, str],
    input_template: str,
    output_template: str,
    n_row_groups_per_partition: int | None,
    row_group_size: int | None,
    verbose: bool | int,
    skip_existing: bool,
    columns_in_partition_names: bool,
) -> None:
    if temporary_dir is None:
        temporary_dir = tempfile.mkdtemp()
    chunks = range_utils.split(all_files, items_per_split=chunk_size)

    # print summary
    if verbose:
        print(
            'Repartitioning',
            len(all_files),
            'files into',
            n_partitions,
            'files',
        )
        total_bytes = sum(os.stat(path).st_size for path in all_files)
        print('- total data:', toolstr.format_nbytes(total_bytes))
        if n_row_groups_per_partition is not None:
            print('- n_row_groups_per_partition:', n_row_groups_per_partition)
        elif row_group_size is not None:
            print('- row_group_size:', row_group_size)
        else:
            print('- using 1 row group per partition')
        original_dirs = {os.path.dirname(path) for path in all_files}
        if len(original_dirs) == 1:
            original_dir = next(iter(original_dirs))
        else:
            original_dir = '[multiple]'
        print()
        print('Directories')
        print('- original dir:', original_dir)
        print('- temporary dir:', temporary_dir)
        print('- final dir:', output_dir)
        print()
        print(
            'partitioning by column',
            column,
            'using',
            len(labels),
            'value ranges:',
        )
        labels_with_number: list[tuple[int | None, str | None]] = list(
            enumerate(labels)
        )
        show_partitions = 10
        if len(labels) > show_partitions:
            labels_with_number = (
                labels_with_number[: int(show_partitions / 2)]
                + [(None, None)]
                + labels_with_number[int(-show_partitions / 2) :]
            )
        for p, label in labels_with_number:
            if p is not None:
                print('    ' + str(p + 1) + '.', label)
            else:
                print('    ...')
        print()
        print('creating', len(chunks) * n_partitions, 'temporary files')
        if len(chunks) == 1:
            word = 'chunk'
        else:
            word = 'chunks'
        print(
            '- processing the original',
            len(all_files),
            'files in',
            len(chunks),
            word,
            'of',
            len(chunks[0]),
            'files each',
        )

    # convert each chunk into parititons
    paths_per_partition: typing.MutableMapping[str, list[str]] = {}
    for c, chunk in enumerate(chunks):
        if verbose:
            aggregations = toolstr.parse_aggregate_by_template(
                strs=chunk,
                template=input_template,
                aggregations=template_aggregations,
            )
            name = ', '.join(
                k + '=' + str(aggregations[k])
                for k in template_aggregations.keys()
            )
            print('processing chunk', c + 1, '/', str(len(chunks)) + ',', name)

        # get output_template
        partition_output_template = _combine_paths(
            paths=chunk,
            input_template=input_template,
            output_template=output_template,
            aggregations=template_aggregations,
            values={'partition': '{partition}'},
        )
        partition_output_template = os.path.join(
            temporary_dir,
            partition_output_template,
        )

        # perform partitioning
        chunk_paths_per_partition = _files_to_partitions(
            paths=chunk,
            bounds=bounds,
            labels=labels,
            column=column,
            output_dir=temporary_dir,
            input_template=input_template,
            output_template=partition_output_template,
            skip_existing=skip_existing,
            columns_in_partition_names=columns_in_partition_names,
        )

        # record file paths
        for key, value in chunk_paths_per_partition.items():
            paths_per_partition.setdefault(key, [])
            paths_per_partition[key].extend(value)

    # consolidate files of each partition
    if verbose:
        print()
        print('consolidating into final partitioned files')
    for p, partition in enumerate(paths_per_partition.keys()):
        partition_paths = paths_per_partition[partition]

        # get output_path
        output_path = _combine_paths(
            paths=partition_paths,
            input_template=output_template,
            output_template=output_template,
            aggregations=template_aggregations,
            values={},
        )
        output_path = os.path.join(output_dir, output_path)

        if skip_existing and os.path.isfile(output_path):
            if verbose:
                print(
                    'skipping',
                    partition,
                    '(partition',
                    str(p + 1) + ' / ' + str(len(labels)) + ')',
                )
            continue
        if verbose >= 2:
            print(
                'writing',
                partition,
                '(partition',
                str(p + 1) + ' / ' + str(len(labels)) + ')',
            )

        # load files and consolidate into single file
        df = pl.concat(
            [pl.read_parquet(path) for path in partition_paths]
        ).sort(column)
        pl_utils.write_df(
            df=df,
            path=output_path,
            n_row_groups=n_row_groups_per_partition,
            row_group_size=row_group_size,
            create_dir=True,
        )

    # delete temporary files
    if delete_temporary:
        if verbose:
            print()
            print('deleting temporary files')
        for partition, partition_paths in paths_per_partition.items():
            for path in partition_paths:
                os.remove(path)
    else:
        if verbose:
            print()
            print('not deleting temporary files')

    # print outro
    if verbose:
        print('reindexing complete')


def _files_to_partitions(
    *,
    paths: typing.Sequence[str],
    bounds: typing.Sequence[int],
    labels: typing.Sequence[str],
    column: str,
    output_dir: str,
    input_template: str,
    output_template: str,
    skip_existing: bool,
    columns_in_partition_names: bool,
) -> typing.MutableMapping[str, typing.Sequence[str]]:
    # create output paths
    paths_per_partition: typing.MutableMapping[str, typing.Sequence[str]] = {}
    for p, label in enumerate(labels):
        if columns_in_partition_names:
            partition_name = column + '_' + label
        else:
            partition_name = label
        paths_per_partition[label] = [
            output_template.format(partition=partition_name)
        ]

    # skip if paths already exist
    if skip_existing and all(
        os.path.isfile(path)
        for path_list in paths_per_partition.values()
        for path in path_list
    ):
        return paths_per_partition

    # load files
    df = pl.concat([pl.scan_parquet(path) for path in paths]).collect(
        streaming=True
    )

    # create cut labels
    n_bound_nibbles = len(labels[-1].split('_')[0]) - 2
    partition_values = (
        df[column]
        .bin.encode('hex')
        .str.slice(0, n_bound_nibbles)
        .apply(lambda x: int(x, 16))
    )

    # determine partition of each row
    binned_rows = partition_values.cut(
        bins=bounds,  # type: ignore
        labels=['UNDERFLOW'] + labels + ['OVERFLOW'],  # type: ignore
        maintain_order=True,
    )
    partitions = binned_rows['category'].cast(int)

    # write partitioned files, and record path of each partition
    for p, label in enumerate(labels):
        output_path = paths_per_partition[label][0]
        partition_df = df.filter(partitions == p + 1)
        pl_utils.write_df(
            df=partition_df,
            path=output_path,
            statistics=False,
            create_dir=True,
        )

    return paths_per_partition


def _combine_paths(
    *,
    paths: typing.Sequence[str],
    input_template: str,
    output_template: str,
    aggregations: typing.Mapping[str, str],
    values: typing.Mapping[str, str],
) -> str:
    """metadata should be separated by underscores"""

    output_values = toolstr.parse_aggregate_by_template(
        strs=paths,
        template=input_template,
        aggregations=aggregations,
    )

    if values is not None:
        output_values = dict(output_values, **values)

    return output_template.format(**output_values)




import pytest
from ctc import evm

rlp_examples = [
    #
    # integers
    [0, '0x80'],
    [15, '0x0f'],
    [1024, '0x820400'],
    #
    # str
    [
        'dog',
        evm.to_hex(bytes.fromhex('83') + 'dog'.encode()),
    ],
    [
        ['cat', 'dog'],
        evm.to_hex(
            bytes.fromhex('c883')
            + 'cat'.encode()
            + bytes.fromhex('83')
            + 'dog'.encode()
        ),
    ],
    #
    # lists
    [[], '0xc0'],
    [
        [[], [[]], [[], [[]]]],
        '0xc7c0c1c0c3c0c1c0',
    ],
]


@pytest.mark.parametrize('test', rlp_examples)
def test_rlp_encoding(test):
    data, target_encoding = test
    actual_encoding = evm.rlp_encode(data, str_mode='text')
    assert actual_encoding == target_encoding


#
# # test decoding
#

ascii_examples = [
    'cat',
    'dog',
    'a',
    'hello',
    'world',
    ['hello', 'world'],
    ['nested', ['list', ['structure']]],
    [],
    [[], [[], []]],
]

raw_hex_examples = [
    '636174',
    ['636174'],
    [],
    [[], [[], []]],
]

prefix_hex_examples = [
    '0x636174',
    ['0x636174'],
    [],
    [[], [[], []]],
]


@pytest.mark.parametrize('test', ascii_examples)
def test_rlp_decoding_ascii(test):
    encoded = evm.rlp_encode(test)
    decoded = evm.rlp_decode(encoded, types='ascii')
    assert decoded == test


@pytest.mark.parametrize('test', raw_hex_examples)
def test_rlp_decoding_raw_hex(test):
    encoded = evm.rlp_encode(test, str_mode='hex')
    decoded = evm.rlp_decode(encoded, types='raw_hex')
    assert decoded == test


@pytest.mark.parametrize('test', prefix_hex_examples)
def test_rlp_decoding_prefix_hex(test):
    encoded = evm.rlp_encode(test)
    decoded = evm.rlp_decode(encoded, types='prefix_hex')
    assert decoded == test


#
# # address nonce tuples
#

nonces = [
    0,
    1,
    2,
    3,
    4,
    126,
    127,
    128,
    129,
    256,
    512,
    1024,
    2048,
    65000,
    66000,
    1320000000,
    13200000000000,
]


encoded_ints = [
    '0x80',
    '0x01',
    '0x02',
    '0x03',
    '0x04',
    '0x7e',
    '0x7f',
    '0x8180',
    '0x8181',
    '0x820100',
    '0x820200',
    '0x820400',
    '0x820800',
    '0x82fde8',
    '0x830101d0',
    '0x844ead9a00',
    '0x860c015d4fa000',
]


address = '0x956f47f50a910163d8bf957cf5846d573e7f87ca'
encoded_address = '0x94956f47f50a910163d8bf957cf5846d573e7f87ca'


encoded_address_nonce_tuples = [
    '0xd694956f47f50a910163d8bf957cf5846d573e7f87ca80',
    '0xd694956f47f50a910163d8bf957cf5846d573e7f87ca01',
    '0xd694956f47f50a910163d8bf957cf5846d573e7f87ca02',
    '0xd694956f47f50a910163d8bf957cf5846d573e7f87ca03',
    '0xd694956f47f50a910163d8bf957cf5846d573e7f87ca04',
    '0xd694956f47f50a910163d8bf957cf5846d573e7f87ca7e',
    '0xd694956f47f50a910163d8bf957cf5846d573e7f87ca7f',
    '0xd794956f47f50a910163d8bf957cf5846d573e7f87ca8180',
    '0xd794956f47f50a910163d8bf957cf5846d573e7f87ca8181',
    '0xd894956f47f50a910163d8bf957cf5846d573e7f87ca820100',
    '0xd894956f47f50a910163d8bf957cf5846d573e7f87ca820200',
    '0xd894956f47f50a910163d8bf957cf5846d573e7f87ca820400',
    '0xd894956f47f50a910163d8bf957cf5846d573e7f87ca820800',
    '0xd894956f47f50a910163d8bf957cf5846d573e7f87ca82fde8',
    '0xd994956f47f50a910163d8bf957cf5846d573e7f87ca830101d0',
    '0xda94956f47f50a910163d8bf957cf5846d573e7f87ca844ead9a00',
    '0xdc94956f47f50a910163d8bf957cf5846d573e7f87ca860c015d4fa000',
]


@pytest.mark.parametrize('test', zip(nonces, encoded_ints))
def test_rlp_encode_integer(test):
    integer, target_encoding = test
    assert evm.rlp_encode(integer) == target_encoding


def test_rlp_encode_address():
    assert evm.rlp_encode(address, str_mode='hex') == encoded_address


@pytest.mark.parametrize('test', zip(nonces, encoded_address_nonce_tuples))
def test_rlp_encode_address_integer_tuples(test):
    nonce, target_encoding = test

    assert evm.rlp_encode([address, nonce], str_mode='hex') == target_encoding




from __future__ import annotations

import typing

from ctc import spec
from .. import dex_class
from .. import dex_class_utils

if typing.TYPE_CHECKING:
    from typing_extensions import Literal

    import tooltime


async def async_get_pool_trades(
    pool: spec.Address,
    *,
    dex: typing.Type[dex_class.DEX] | str | None = None,
    factory: spec.Address | None = None,
    normalize: bool = False,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    label: Literal['index', 'symbol', 'address'] = 'index',
    include_timestamps: bool = False,
    remove_missing_fields: bool = True,
    include_prices: bool = False,
    include_volumes: bool = False,
    context: spec.Context = None,
) -> spec.DataFrame:
    """get trades of a DEX pool"""

    dex = await dex_class_utils.async_get_dex_class(
        dex=dex,
        factory=factory,
        pool=pool,
        context=context,
    )

    return await dex.async_get_pool_trades(
        pool=pool,
        normalize=normalize,
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        label=label,
        context=context,
        include_timestamps=include_timestamps,
        remove_missing_fields=remove_missing_fields,
        include_prices=include_prices,
        include_volumes=include_volumes,
    )



"""
see also
    https://meta.yearn.finance/api/1/vaults/all
    https://meta.yearn.finance/api/1/protocols/all
    https://meta.yearn.finance/api/1/tokens/all
    https://meta.yearn.finance/api/1/strategies/all
"""

from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from . import yearn_spec


vaults_endpoint = 'https://api.yearn.finance/v1/chains/{chain_id}/vaults/all'


async def async_get_yearn_api_vaults(
    network: spec.NetworkReference,
) -> typing.Sequence[yearn_spec.ApiVault]:
    import aiohttp

    chain_id = evm.get_network_chain_id(network)
    url = vaults_endpoint.format(chain_id=chain_id)

    async with aiohttp.ClientSession() as session:
        async with session.get(url) as response:
            vaults: typing.Sequence[yearn_spec.ApiVault] = await response.json()
            return vaults


async def async_get_yearn_api_vault(
    query: str,
    network: spec.NetworkReference,
    *,
    api_vaults: typing.Sequence[typing.Any] | None = None,
) -> yearn_spec.ApiVault:
    """

    Query will return the first of the following that matches:
    - query is an address of a vault
    - query equals the name of a vault
    - query equals the symbol of a vault (result whichever has latest version)

    All queries are considered case insensitive
    """

    # obtain api data
    if api_vaults is None:
        api_vaults = await async_get_yearn_api_vaults(network=network)

    # check if query matches an address
    query = query.lower()
    if evm.is_address_str(query):
        for vault in api_vaults:
            if vault['address'].lower() == query:
                return vault

    # check if query matches a name
    for vault in api_vaults:
        if query == vault['name'].lower():
            return vault

    # check if query matches a symbol
    candidates = []
    for vault in api_vaults:
        if query == vault['symbol'].lower():
            candidates.append(vault)
    if len(candidates) == 0:
        raise Exception('vault could not be found')
    elif len(candidates) == 1:
        return candidates[0]
    else:
        # TODO: more robust determination of latest vault
        ordered = sorted(
            candidates, key=lambda candidate: candidate['name']
        )
        latest = ordered[-1]
        return latest



from __future__ import annotations

from ctc import spec
from .. import event_utils
from . import erc721_spec


async def async_get_erc721_transfers(
    token: spec.Address,
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> spec.DataFrame:
    """get transfers of erc721 token"""
    return await event_utils.async_get_events(
        contract_address=token,
        event_abi=erc721_spec.erc721_event_abis['Transfer'],
        verbose=False,
        start_block=start_block,
        end_block=end_block,
        context=context,
    )


async def async_get_erc721_approvals(
    token: spec.Address,
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> spec.DataFrame:
    """get approval of erc721 token"""
    return await event_utils.async_get_events(
        contract_address=token,
        event_abi=erc721_spec.erc721_event_abis['Approve'],
        verbose=False,
        start_block=start_block,
        end_block=end_block,
        context=context,
    )


async def async_get_erc721_approvals_for_all(
    token: spec.Address,
    *,
    start_block: spec.BlockNumberReference | None = None,
    end_block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> spec.DataFrame:
    """get approvals for all of erc721 token"""
    return await event_utils.async_get_events(
        contract_address=token,
        event_abi=erc721_spec.erc721_event_abis['ApproveForAll'],
        verbose=False,
        start_block=start_block,
        end_block=end_block,
        context=context,
    )



from __future__ import annotations

import toolcli

from ctc import evm


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_symbol_command,
        'help': 'convert ERC20 address to symbol, or symbol to address',
        'args': [
            {'name': 'query', 'help': 'ERC20 address or symbol'},
        ],
        'examples': [
            'DAI',
        ],
    }


async def async_symbol_command(query: str) -> None:
    if evm.is_address_str(query):
        symbol = await evm.async_get_erc20_symbol(query)
        print(symbol)
    else:
        try:
            address = await evm.async_get_erc20_address(query)
            print(address)
        except Exception:
            print(
                'could not find address for token with symbol "'
                + str(query)
                + '"'
            )



"""
TODO:
- add grouping for specific groups:
        - erc20 functions and events
        - proxy functions
        - erc721 erc1155
    - could group these either spatially or by color
- add listing of unknown signatures
- add event signatures
"""
from __future__ import annotations

import typing

import toolcli
import toolstr

from ctc import cli
from ctc import evm
from ctc import rpc
from ctc import spec


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_decompile_command,
        'help': 'decompile contract abi',
        'args': [
            {
                'name': 'address_or_bytecode',
                'help': 'contract address or hex bytecode',
            },
            {
                'name': ['-v', '--verbose'],
                'help': 'do not clip long names',
                'action': 'store_true',
            },
        ],
        'examples': [
            '0x956f47f50a910163d8bf957cf5846d573e7f87ca',
        ],
    }


async def async_decompile_command(
    address_or_bytecode: str,
    verbose: bool,
    *,
    context: spec.Context = None,
) -> None:

    address_or_bytecode = await evm.async_resolve_address(address_or_bytecode)

    if len(address_or_bytecode) == 42:
        bytecode = await rpc.async_eth_get_code(address_or_bytecode)
    else:
        bytecode = address_or_bytecode

    # get function selectors
    function_selectors = evm.extract_bytecode_function_selectors(bytecode)
    function_selector_indices = {
        function_selector: index
        for index, function_selector in enumerate(sorted(function_selectors))
    }

    # match against 4bytes
    decompiled_function_abis = await evm.async_decompile_function_abis(
        bytecode,
        sort='hex_signature',
        context=context,
    )

    styles = cli.get_cli_styles()
    toolstr.print(
        'Found '
        + toolstr.add_style(
            str(len(function_selectors)),
            styles['description'] + ' bold',
        )
        + ' function selectors matching '
        + toolstr.add_style(
            str(len(decompiled_function_abis)),
            styles['description'] + ' bold',
        )
        + ' total 4byte entries',
    )

    if len(decompiled_function_abis) > 0:
        print()
        toolstr.print_header('Known selectors', style=styles['title'])
    rows: typing.MutableSequence[typing.MutableSequence[typing.Any]] = []
    row = None
    for e, entry in enumerate(decompiled_function_abis):

        duplicate = (
            e > 0
            and entry['hex_signature']
            == decompiled_function_abis[e - 1]['hex_signature']
        )

        if duplicate and row is not None:
            row[2] = row[2] + '\n' + entry['text_signature']
            # row = [
            #     '',
            #     '',
            #     entry['text_signature'],
            # ]
        else:
            if row is not None:
                rows.append(row)
            row = [
                function_selector_indices[entry['hex_signature']],
                entry['hex_signature'],
                entry['text_signature'],
            ]
    if verbose:
        width = None
    else:
        width = toolcli.get_n_terminal_cols()

    print()
    toolstr.print_multiline_table(
        rows,
        column_justify=['right', 'right', 'left'],
        # add_row_index=True,
        labels=['', 'selector', '4byte matches'],
        border=styles['comment'],
        label_style=styles['title'],
        column_styles=[
            styles['title'] + ' bold',
            styles['metavar'],
            styles['description'],
        ],
        # compact=2,
        max_table_width=width,
        # indent=4,
    )

    decompiled_selectors = set(
        entry['hex_signature'] for entry in decompiled_function_abis
    )
    unknown_selectors = set(function_selectors) - decompiled_selectors
    if len(unknown_selectors) > 0:
        print()
        toolstr.print_header('Unknown selectors', style=styles['title'])
        for unknown_selector in unknown_selectors:
            print(unknown_selector)

    if len(decompiled_function_abis) == 0:
        print()
        print('could not detect any function signatures')




from __future__ import annotations

from ctc import spec
from . import context_caches
from . import context_sources


def create_user_input_context(
    *,
    provider: str | None = None,
    network: str | spec.NetworkReference | None = None,
    cache: spec.ContextCacheShorthand = None,
) -> spec.Context:

    if isinstance(network, str) and network.isnumeric():
        network = int(network)

    context: spec.ShorthandContext = {
        'provider': provider,
        'network': network,
        'cache': cache,
    }

    return context


def normalize_context(context: spec.Context) -> spec.NormalizedContext:
    network, provider = context_sources._get_context_chain_id_and_provider(
        context
    )
    cache_rules = context_caches._extract_context_cache_rules(context=context)
    return {
        'network': network,
        'provider': provider,
        'cache': cache_rules,
    }


def update_context(
    context: spec.Context,
    *,
    network: spec.NetworkReference | None = None,
    provider: spec.ProviderReference | None = None,
    merge_provider: spec.PartialProvider | None = None,
    cache: spec.ContextCacheShorthand = None,
) -> spec.NormalizedContext:

    new_context = normalize_context(context)

    if network is not None:
        new_context['network'] = network
    if provider is not None:
        new_context['provider'] = provider
    if merge_provider is not None:
        from ctc import rpc

        if provider is None:
            new_context['provider'] = rpc.resolve_provider(merge_provider)
        else:
            provider = rpc.resolve_provider(provider)
            new_context['provider'] = dict(provider, **merge_provider)  # type: ignore

    if cache is not None:
        cache_rules = context_caches._extract_context_cache_rules(
            context_cache=cache
        )
        new_context['cache'] = list(cache_rules) + list(new_context['cache'])

    return new_context




from __future__ import annotations

import typing
from typing_extensions import TypedDict, NotRequired

from . import external_types


class RawDexTrades(TypedDict):
    block_number: external_types.Series
    timestamp: NotRequired[external_types.Series | typing.Sequence[int] | None]
    transaction_hash: external_types.Series
    recipient: NotRequired[external_types.Series | None]
    bought_id: external_types.Series
    sold_id: external_types.Series
    bought_amount: external_types.Series
    sold_amount: external_types.Series



from __future__ import annotations

import typing

from ctc import spec
from . import erc4626_spec


async def async_get_erc4626_asset(
    token: spec.Address,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> spec.Address:
    """get asset address of ERC-4626 vault"""
    from ctc import rpc

    asset: spec.Address = await rpc.async_eth_call(
        to_address=token,
        context=context,
        function_abi=erc4626_spec.erc4626_function_abis['asset'],
        block_number=block,
    )
    return asset


async def async_get_erc4626s_assets(
    tokens: typing.Sequence[spec.Address],
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> typing.Sequence[spec.Address]:
    """get asset addresses of ERC-4626 vaults"""
    from ctc import rpc

    assets: typing.Sequence[spec.Address] = await rpc.async_batch_eth_call(
        to_addresses=tokens,
        context=context,
        function_abi=erc4626_spec.erc4626_function_abis['asset'],
        block_number=block,
    )
    return assets



import inspect
import os
import sys
import types

import ctc
import ctc.db
import ctc.rpc


def test_ctc_functions_have_docstrings():
    functiontype = type(test_ctc_functions_have_docstrings)

    functions_without_docstrings = []
    for key, value in vars(ctc).items():
        if isinstance(value, functiontype):
            docstring = value.__doc__
            missing = False
            if docstring is None or docstring == '':
                missing = True
            else:
                lines = docstring.split('\n')
                if lines[0] == '':
                    missing = True

            if missing:
                name = value.__name__
                file = sys.modules[value.__module__].__file__
                functions_without_docstrings.append((name, file))

    if len(functions_without_docstrings) > 0:
        message = 'some functions do not have docstrings:\n'
        for name, file in functions_without_docstrings:
            message += '    - ' + name + ': ' + file + '\n\n'
        raise Exception(message)


def test_code_imports_future_annotations():
    phrase = 'from __future__ import annotations'
    root_path = ctc.__path__[0]

    missing = []
    for root, dirnames, filenames in os.walk(root_path):
        for filename in filenames:
            if not filename.startswith('__') and filename.endswith('.py'):
                filepath = os.path.join(root, filename)
                with open(filepath, 'r') as f:
                    content = f.read()
                if phrase not in content:
                    missing.append(filepath)

    if len(missing) > 0:
        relpaths = [os.path.relpath(path, root_path) for path in missing]
        print()
        print()
        print(len(missing), 'paths with missing future annotation import:')
        print()
        print('- ' + '\n- '.join(relpaths))
    assert len(missing) == 0


def test_every_module_has_init_file():
    root_path = ctc.__path__[0]

    missing = []
    for root, dirnames, filenames in os.walk(root_path):
        needs_init = any(filename.endswith('.py') for filename in filenames)

        if needs_init:
            if '__init__.py' not in os.listdir(root):
                missing.append(root)

    if len(missing) > 0:
        relpaths = [os.path.relpath(path, root_path) for path in missing]
        print()
        print()
        print(len(missing), 'modules with missing __init__:')
        print()
        print(' '.join(relpaths))
    assert len(missing) == 0


def test_function_names():
    import importlib
    import inspect
    import pkgutil

    should_have_async_in_name = []
    should_not_have_async_in_name = []
    async_exceptions = [
        '__aenter__',
        '__aexit__',
        'async_connect',
    ]

    should_use_by_block = []
    by_block_exceptions = [
        'bin_by_blocks',
    ]

    for importer, modname, ispkg in pkgutil.walk_packages(
        path=ctc.__path__,
        prefix="ctc.",
        onerror=lambda x: None,
    ):
        # skip files that run a lot of code
        if modname.endswith('__main__'):
            continue

        module = importlib.import_module(modname)
        for attr_name in dir(module):
            module_attr = getattr(module, attr_name)
            if hasattr(module_attr, '__call__'):
                check_function_name(
                    module_attr=module_attr,
                    attr_name=attr_name,
                    modname=modname,
                    should_have_async_in_name=should_have_async_in_name,
                    should_not_have_async_in_name=should_not_have_async_in_name,
                    should_use_by_block=should_use_by_block,
                    async_exceptions=async_exceptions,
                    by_block_exceptions=by_block_exceptions,
                )

            # check methods of classes
            if (
                isinstance(module_attr, type)
                and module_attr.__module__ == modname
            ):
                methods = inspect.getmembers(
                    module_attr, predicate=inspect.ismethod
                )
                for method_name, method in methods:
                    check_function_name(
                        module_attr=method,
                        attr_name=method_name,
                        modname=modname + '.' + attr_name,
                        should_have_async_in_name=should_have_async_in_name,
                        should_not_have_async_in_name=should_not_have_async_in_name,
                        should_use_by_block=should_use_by_block,
                        async_exceptions=async_exceptions,
                        by_block_exceptions=by_block_exceptions,
                    )

    if len(should_have_async_in_name) > 0:
        message = 'names of these items should start with async_ or _async_:'
        for attr in sorted(should_have_async_in_name):
            message += '\n    - ' + attr
        raise Exception(message)
    if len(should_not_have_async_in_name) > 0:
        message = (
            'names of these items should not start with async_ or _async_:'
        )
        for attr in sorted(should_not_have_async_in_name):
            message += '\n    - ' + attr
        raise Exception(message)
    if len(should_use_by_block) > 0:
        message = (
            'functions should use by_block instead of per_block or by_blocks'
        )
        for attr in sorted(should_use_by_block):
            message += '\n    - ' + attr
        raise Exception(message)


def check_function_name(
    module_attr,
    attr_name,
    modname,
    should_have_async_in_name,
    should_not_have_async_in_name,
    should_use_by_block,
    async_exceptions,
    by_block_exceptions,
):
    import inspect

    named_as_async = attr_name.startswith('async') or attr_name.startswith(
        '_async'
    )
    if inspect.iscoroutinefunction(module_attr):
        if not named_as_async and attr_name not in async_exceptions:
            should_have_async_in_name.append(modname + '.' + attr_name)
    else:
        if named_as_async and attr_name not in async_exceptions:
            should_not_have_async_in_name.append(modname + '.' + attr_name)

    if 'per_block' in attr_name or 'by_blocks' in attr_name:
        if attr_name not in by_block_exceptions:
            should_use_by_block.append(modname + '.' + attr_name)


def iterate_package_functions(prefix):
    import importlib
    import pkgutil

    functions = {}
    function_to_name = {}

    for importer, modname, ispkg in pkgutil.walk_packages(
        path=ctc.__path__,
        prefix=prefix,
        onerror=lambda x: None,
    ):
        # skip files that run a lot of code
        if modname.endswith('__main__'):
            continue

        module = importlib.import_module(modname)
        for attr_name in dir(module):
            module_attr = getattr(module, attr_name)
            if isinstance(module_attr, types.FunctionType):
                function_data = {
                    'module_name': modname,
                    'function_name': attr_name,
                    'function': module_attr,
                    'module': module,
                }
                name = modname + '.' + attr_name

                if module_attr not in function_to_name:
                    functions[name] = function_data
                    function_to_name[module_attr] = name
                else:
                    # use longest valid name for function
                    other_name = function_to_name[module_attr]
                    if len(name) > len(other_name):
                        del functions[other_name]
                        functions[name] = function_data
                        function_to_name[module_attr] = name

    return functions


def test_max_positional_args():
    import inspect

    max_positional_args = 2
    allowed_exceptions = [
        'ctc.toolbox.chunk_utils.range_to_chunks',
    ]

    too_many = []
    for function_name, function_data in iterate_package_functions(
        'ctc.'
    ).items():
        argspec = inspect.getfullargspec(function_data['function'])
        if len(argspec.args) > max_positional_args:
            if function_name not in allowed_exceptions:
                too_many.append(function_name)

    for class_name, class_data in iterate_package_classes('ctc.').items():
        for method_name, method in class_data['methods']:
            argspec = inspect.getfullargspec(method)
            if len(argspec.args) > max_positional_args + 1:
                name = class_name + '.' + method_name
                if name not in allowed_exceptions:
                    too_many.append(name)

    if len(too_many) > 0:
        message = (
            'functions should take no more than '
            + str(max_positional_args)
            + ' positional args, use `*` to convert to keyword-only arguments'
        )
        for item in too_many:
            message += '\n- ' + item
        message += '\n' + str(len(too_many)) + ' violations'
        message += '\n\nuse `*` to convert to keyword-only arguments'
        raise Exception(message)


def iterate_package_classes(prefix, include_methods=True):
    import importlib
    import inspect
    import pkgutil

    classes = {}
    classes_to_name = {}

    for importer, modname, ispkg in pkgutil.walk_packages(
        path=ctc.__path__,
        prefix=prefix,
        onerror=lambda x: None,
    ):
        # skip files that run a lot of code
        if modname.endswith('__main__'):
            continue

        module = importlib.import_module(modname)
        for attr_name in dir(module):
            module_attr = getattr(module, attr_name)
            if (
                isinstance(module_attr, type)
                and module_attr.__module__ == module.__name__
            ):
                class_data = {
                    'module_name': modname,
                    'class_name': attr_name,
                    'class': module_attr,
                    'module': module,
                }

                if include_methods:
                    class_data['methods'] = inspect.getmembers(
                        module_attr,
                        predicate=inspect.ismethod,
                    )

                name = modname + '.' + attr_name
                if module_attr not in classes_to_name:
                    classes[name] = class_data
                    classes_to_name[module_attr] = name
                else:
                    # use longest valid name for function
                    other_name = classes_to_name[module_attr]
                    if len(name) > len(other_name):
                        del classes[other_name]
                        classes[name] = class_data
                        classes_to_name[module_attr] = name

    return classes


def _get_tested_modules():
    from ctc.protocols import (
        aave_v2_utils,
        balancer_utils,
        chainlink_utils,
        coingecko_utils,
        compound_utils,
        curve_utils,
        ens_utils,
        etherscan_utils,
        fourbyte_utils,
        g_uni_utils,
        gnosis_utils,
        llama_utils,
        multicall_utils,
        rari_utils,
        sushi_utils,
        uniswap_v2_utils,
        uniswap_v3_utils,
        yearn_utils,
    )

    return [
        ctc,
        ctc.db,
        ctc.rpc,
        aave_v2_utils,
        balancer_utils,
        chainlink_utils,
        coingecko_utils,
        compound_utils,
        curve_utils,
        ens_utils,
        etherscan_utils,
        fourbyte_utils,
        g_uni_utils,
        gnosis_utils,
        llama_utils,
        multicall_utils,
        rari_utils,
        sushi_utils,
        uniswap_v2_utils,
        uniswap_v3_utils,
        yearn_utils,
    ]


def _get_take_context_exceptions():
    from ctc.protocols import uniswap_v3_utils

    return [
        (ctc.rpc, 'async_send_raw'),
        (ctc.rpc, 'async_close_http_session'),
        (uniswap_v3_utils, 'async_get_function_abi'),
        (uniswap_v3_utils, 'async_get_event_abi'),
    ]


def test_async_functions_take_context():
    modules = _get_tested_modules()
    exceptions = _get_take_context_exceptions()

    failures = []
    for module in modules:
        for name, value in vars(module).items():
            if name.startswith('async_') and isinstance(
                value, types.FunctionType
            ):
                # try exception list first
                if module == ctc.rpc and name.startswith('async_batch_'):
                    continue
                elif module.__name__ in (
                    'ctc.protocols.coingecko_utils',
                    'ctc.protocols.fourbyte_utils',
                    'ctc.protocols.llama_utils',
                    'ctc.protocols.rari_utils',
                    'ctc.protocols.yearn_utils',
                ):
                    continue
                elif (module, name) in exceptions:
                    continue

                argspec = inspect.getfullargspec(value)
                if (
                    'context' not in argspec.args
                    and 'context' not in argspec.kwonlyargs
                ):
                    failures.append(value.__module__ + '.' + name)
    if len(failures) > 0:
        raise Exception(
            str(len(failures))
            + ' functions missing context argument:\n- '
            + '\n- '.join(failures)
        )


def _get_take_kw_only_context_exceptions():
    return [
        (ctc.rpc, 'async_send_raw'),
        (ctc.rpc, 'async_close_http_session'),
        (ctc, 'async_get_latest_block_number'),
        (ctc, 'async_get_default_erc20_tokens'),
    ]


def test_async_functions_take_kw_only_context():
    modules = _get_tested_modules()
    exceptions = _get_take_kw_only_context_exceptions()

    failures = []
    for module in modules:
        for name, value in vars(module).items():

            if name.startswith('async_') and isinstance(
                value, types.FunctionType
            ):
                # try exception list first
                if module == ctc.rpc and name.startswith('async_batch_'):
                    continue
                elif module.__name__ in (
                    'ctc.protocols.coingecko_utils',
                    'ctc.protocols.fourbyte_utils',
                    'ctc.protocols.llama_utils',
                    'ctc.protocols.rari_utils',
                ):
                    continue
                elif (module, name) in exceptions:
                    continue

                argspec = inspect.getfullargspec(value)
                if 'context' in argspec.args:

                    # skip ctc.db.connect_utils.async_connect()
                    if (
                        value.__module__ == 'ctc.db.connect_utils'
                        and name == 'async_connect'
                    ):
                        continue

                    failures.append(value.__module__ + '.' + name)

    if len(failures) > 0:
        raise Exception(
            str(len(failures))
            + ' functions that use context as a positional argument:\n- '
            + '\n- '.join(failures)
        )




from __future__ import annotations

from ctc import rpc
from ctc import spec

from .. import uniswap_v3_spec


async def async_quote_exact_input_single(
    token_in: spec.Address,
    token_out: spec.Address,
    *,
    fee: int,
    amount_in: int,
    sqrt_price_limit_x96: int = 0,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'quoteExactInputSingle',
        'quoter',
    )
    result = await rpc.async_eth_call(
        to_address=uniswap_v3_spec.quoter,
        function_abi=function_abi,
        function_parameters=[
            token_in,
            token_out,
            fee,
            amount_in,
            sqrt_price_limit_x96,
        ],
        context=context,
        block_number=block,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_quote_exact_input(
    path: str,
    amount_in: int,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'quoteExactInput',
        'quoter',
    )
    result = await rpc.async_eth_call(
        to_address=uniswap_v3_spec.quoter,
        function_abi=function_abi,
        function_parameters=[path, amount_in],
        context=context,
        block_number=block,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_quote_exact_output_single(
    token_in: spec.Address,
    token_out: spec.Address,
    *,
    fee: int,
    amount_out: int,
    sqrt_price_limit_x96: int = 0,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'quoteExactOutputSingle',
        'quoter',
    )
    result = await rpc.async_eth_call(
        to_address=uniswap_v3_spec.quoter,
        function_abi=function_abi,
        function_parameters=[
            token_in,
            token_out,
            fee,
            amount_out,
            sqrt_price_limit_x96,
        ],
        context=context,
        block_number=block,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_quote_exact_output(
    path: str,
    amount_in: int,
    *,
    context: spec.Context = None,
    block: spec.BlockNumberReference | None = None,
) -> int:
    function_abi = await uniswap_v3_spec.async_get_function_abi(
        'quoteExactOutput',
        'quoter',
    )
    result = await rpc.async_eth_call(
        to_address=uniswap_v3_spec.quoter,
        function_abi=function_abi,
        function_parameters=[path, amount_in],
        context=context,
        block_number=block,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result




from __future__ import annotations

import asyncio
import typing

from ctc import evm
from ctc import spec

from .. import chainlink_aggregators
from .. import chainlink_feed_metadata
from .. import chainlink_spec
from . import feed_datum

if typing.TYPE_CHECKING:
    import tooltime
    import polars as pl


async def async_get_full_feed_event_data(
    feed: chainlink_spec._FeedReference,
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    normalize: bool = True,
    interpolate: bool = False,
    context: spec.Context = None,
    invert: bool = False,
) -> spec.DataFrame:
    """
    TODO: be able to gather data across multiple aggregator changes
    """

    from ctc.toolbox import pl_utils
    import polars as pl

    # get feed address
    feed = await chainlink_feed_metadata.async_resolve_feed_address(
        feed, context=context
    )

    # get aggregator
    start_block, end_block = await evm.async_resolve_block_range(
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        allow_none=True,
        context=context,
    )
    if start_block is None:
        start_block = await chainlink_feed_metadata.async_get_feed_first_block(
            feed=feed,
            context=context,
        )
    if end_block is None:
        end_block = await evm.async_get_latest_block_number(context=context)
    aggregator_address = (
        await chainlink_feed_metadata.async_get_feed_aggregator(
            feed=feed,
            block=end_block,
            context=context,
        )
    )

    # check if dealing with multiple aggregators
    initial_aggregator_address = (
        await chainlink_feed_metadata.async_get_feed_aggregator(
            feed=feed,
            block=start_block,
            context=context,
        )
    )
    if aggregator_address != initial_aggregator_address:
        history = await chainlink_aggregators.async_get_feed_aggregator_history(
            feed=feed,
            context=context,
        )
        aggregator_starts = list(history.values())
        aggregator_ends = [block - 1 for block in aggregator_starts[1:]]
        start_block = await evm.async_block_number_to_int(
            start_block, context=context
        )
        end_block = await evm.async_block_number_to_int(
            end_block, context=context
        )
        aggregator_ends.append(end_block)
        coroutines = []
        for aggregator, aggregator_start, aggregator_end in zip(
            history.keys(), aggregator_starts, aggregator_ends
        ):
            if aggregator_start < start_block:
                use_start = start_block
            else:
                use_start = aggregator_start

            if aggregator_end > end_block:
                use_end = end_block
            else:
                use_end = aggregator_end

            if aggregator_start > aggregator_end:
                continue

            coroutine = evm.async_get_events(
                contract_address=aggregator,
                event_abi=chainlink_spec.aggregator_event_abis['AnswerUpdated'],
                start_block=use_start,
                end_block=use_end,
                verbose=False,
                context=context,
                integer_output_format={'updatedAt': int},
            )
            coroutines.append(coroutine)
        results = await asyncio.gather(*coroutines)

        # currently need this check as empty results do not have arg columns
        non_empty_results = [result for result in results if len(result) > 0]
        if len(non_empty_results) == 0:
            df: spec.DataFrame = results[0]
        else:
            df = pl_utils.concat(non_empty_results)

    else:

        df = await evm.async_get_events(
            contract_address=aggregator_address,
            event_abi=chainlink_spec.aggregator_event_abis['AnswerUpdated'],
            start_block=start_block,
            end_block=end_block,
            verbose=False,
            context=context,
            integer_output_format={'updatedAt': pl.Int64},
        )

    # rename columns
    new_columns = {
        'arg__current': 'answer',
        'arg__updatedAt': 'timestamp',
        'arg__roundId': 'round_id',
    }
    df = df.rename(new_columns)
    df = df[['block_number', 'answer', 'timestamp', 'round_id']]

    # normalize
    if normalize:
        import numpy as np

        decimals = await chainlink_feed_metadata.async_get_feed_decimals(
            feed=feed, context=context
        )
        df = df.with_columns(
            pl.Series(
                'answer',
                np.array(df['answer'].to_list(), dtype=float) / (10**decimals),
            )
        )

    # interpolate
    if interpolate:

        # TODO: better detection of initial feed data point
        first_feed_block = (
            await chainlink_feed_metadata.async_get_feed_first_block(
                feed, context=context
            )
        )
        if start_block == first_feed_block:
            pass
        else:

            # add initial data
            if len(df) == 0 or start_block < df['block_number'][0]:
                initial_data = await feed_datum.async_get_feed_datum(
                    feed=feed,
                    block=start_block,
                    normalize=normalize,
                    fields='full',
                    context=context,
                )
                schema = {
                    'block_number': pl.Int64,
                    'timestamp': pl.Int64,
                    'round_id': pl.Object,
                }
                if normalize:
                    schema['answer'] = pl.Float64
                else:
                    schema['answer'] = pl.Decimal
                initial_data['block_number'] = start_block
                initial_df = pl.DataFrame(initial_data, schema=schema)
                df = pl_utils.concat([initial_df, df])

        end_block = await evm.async_block_number_to_int(
            end_block, context=context
        )
        df = pl_utils.interpolate(
            df, index_column='block_number', end_index=end_block
        )

    if invert:
        df = df.with_columns(1 / df['answer'])

    return df


async def async_get_answer_feed_event_data(
    feed: chainlink_spec._FeedReference,
    *,
    start_block: typing.Optional[spec.BlockNumberReference] = None,
    end_block: typing.Optional[spec.BlockNumberReference] = None,
    start_time: tooltime.Timestamp | None = None,
    end_time: tooltime.Timestamp | None = None,
    normalize: bool = True,
    interpolate: bool = False,
    invert: bool = False,
    context: spec.Context = None,
) -> spec.DataFrame:

    from ctc.toolbox import pl_utils

    df = await async_get_full_feed_event_data(
        feed=feed,
        start_block=start_block,
        end_block=end_block,
        start_time=start_time,
        end_time=end_time,
        normalize=normalize,
        interpolate=interpolate,
        context=context,
        invert=invert,
    )

    output = df[['block_number', 'answer']]

    return output




from __future__ import annotations

import toolcli

from ctc.protocols import ens_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': hash_command,
        'help': 'output hash of ENS name',
        'args': [
            {'name': 'name', 'help': 'ENS name'},
        ],
        'examples': [
            'vitalik.eth',
        ],
    }


def hash_command(name: str) -> None:
    print(ens_utils.hash_name(name))



from __future__ import annotations

import typing

import ctc
from ctc import spec

if typing.TYPE_CHECKING:
    import tooltime


async def async_compute_block_years(
    block_numbers: typing.Sequence[int] | spec.Series,
) -> spec.Series:
    return await async_compute_block_time_bins(
        block_numbers,
        interval_size='1y',
        alias='year',
    )


async def async_compute_block_months(
    block_numbers: typing.Sequence[int] | spec.Series,
) -> spec.Series:
    return await async_compute_block_time_bins(
        block_numbers,
        interval_size='1M',
        alias='month',
    )


async def async_compute_block_weeks(
    block_numbers: typing.Sequence[int] | spec.Series,
) -> spec.Series:
    return await async_compute_block_time_bins(
        block_numbers,
        interval_size='1w',
        alias='week',
    )


async def async_compute_block_days(
    block_numbers: typing.Sequence[int] | spec.Series,
) -> spec.Series:
    return await async_compute_block_time_bins(
        block_numbers,
        interval_size='1d',
        alias='day',
    )


async def async_compute_block_hours(
    block_numbers: typing.Sequence[int] | spec.Series,
) -> spec.Series:
    return await async_compute_block_time_bins(
        block_numbers,
        interval_size='1h',
        alias='hour',
    )


async def async_compute_block_minutes(
    block_numbers: typing.Sequence[int] | spec.Series,
) -> spec.Series:
    return await async_compute_block_time_bins(
        block_numbers,
        interval_size='1m',
        alias='minute',
    )


async def async_compute_block_time_bins(
    block_numbers: typing.Sequence[int] | spec.Series,
    interval_size: tooltime.Timelength,
    alias: str | None = None,
    context: spec.Context = None,
) -> spec.Series:
    block_intervals = await ctc.async_get_block_intervals(
        start_block=min(block_numbers),
        end_block=max(block_numbers),
        interval_size=interval_size,
        blocks_at='start',
        context={'cache': False},
    )
    return compute_block_time_bins(
        block_numbers,
        block_intervals=block_intervals,
        alias=alias,
    )


def compute_block_time_bins(
    block_numbers: typing.Sequence[int] | spec.Series,
    block_intervals: spec.DataFrame,
    *,
    alias: str | None = None,
) -> spec.Series:
    import polars as pl

    if not isinstance(block_numbers, pl.Series):
        block_numbers = pl.Series(block_numbers)
    if alias is None:
        alias = _compute_block_intervals_alias(block_intervals)
    if alias is None:
        raise Exception('could not determine name of time bin')

    return (
        block_numbers.cut(
            bins=list(block_intervals['block'][1:]),
            labels=list(block_intervals['label']),
        )['category']
        .cast(pl.Utf8)
        .alias(alias)
    )


def _compute_block_intervals_alias(
    block_intervals: spec.DataFrame,
) -> str:

    if len(block_intervals) < 2:
        return None

    timestamps = block_intervals['start_timestamp']
    interval_sizes = set(timestamps[1:] - timestamps[:-1])

    if interval_sizes.issubset({365 * 86400, 366 * 86400}):
        return 'year'
    elif interval_sizes.issubset(
        {28 * 86400, 29 * 86400, 30 * 86400, 31 * 86400}
    ):
        return 'month'
    elif interval_sizes.issubset({86400 * 7}):
        return 'week'
    elif interval_sizes.issubset({86400}):
        return 'day'
    elif interval_sizes.issubset({3600}):
        return 'hour'
    elif interval_sizes.issubset({60}):
        return 'minute'
    elif interval_sizes.issubset({1}):
        return 'second'
    else:
        raise Exception('unknown label for interval')




from __future__ import annotations

import typing

from ctc import evm
from ctc import spec
from ctc.protocols import balancer_utils
from .. import dex_class

if typing.TYPE_CHECKING:
    import tooltime


class BalancerDEX(dex_class.DEX):
    """Balancer DEX"""

    _pool_factories = {1: ['0xba12222222228d8ba445958a75a0704d566bf2c8']}

    @classmethod
    async def async_get_new_pools(
        cls,
        *,
        factory: spec.Address,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.DexPool]:

        balancer_pools = await evm.async_get_events(
            factory,
            event_abi=balancer_utils.vault_event_abis['PoolRegistered'],
            verbose=False,
            start_block=start_block,
            end_block=end_block,
            start_time=start_time,
            end_time=end_time,
            context=context,
        )
        token_registrations = (
            await balancer_utils.async_get_token_registrations(
                factory=factory,
                start_block=start_block,
                end_block=end_block,
                context=context,
            )
        )

        dex_pools = []
        for row in balancer_pools.iter_rows(named=True):

            block = int(row['block_number'])

            assets: typing.Sequence[str | None] = token_registrations.get(
                row['arg__poolId'], []
            )
            if len(assets) < 4:
                assets = list(assets) + [None] * (4 - len(assets))
            if len(assets) > 4:
                additional_data = {'additional_assets': assets[4:]}
                assets = assets[:4]
            else:
                additional_data = {}
            asset0 = assets[0]
            asset1 = assets[1]
            asset2 = assets[2]
            asset3 = assets[3]

            dex_pool: spec.DexPool = {
                'address': row['arg__poolAddress'],
                'factory': factory,
                'asset0': asset0,
                'asset1': asset1,
                'asset2': asset2,
                'asset3': asset3,
                'creation_block': block,
                'fee': None,
                'additional_data': additional_data,
            }
            dex_pools.append(dex_pool)

        return dex_pools

    @classmethod
    async def _async_get_pool_assets_from_node(
        cls,
        pool: spec.Address,
        *,
        block: spec.BlockNumberReference | None = None,
        context: spec.Context = None,
    ) -> typing.Sequence[spec.Address]:

        result = await balancer_utils.async_get_pool_balances(
            pool_address=pool, context=context, block=block
        )

        return list(result.keys())

    @classmethod
    async def _async_get_pool_raw_trades(
        cls,
        pool: spec.Address,
        *,
        start_block: spec.BlockNumberReference | None = None,
        end_block: spec.BlockNumberReference | None = None,
        start_time: tooltime.Timestamp | None = None,
        end_time: tooltime.Timestamp | None = None,
        include_timestamps: bool = False,
        verbose: bool = False,
        context: spec.Context = None,
    ) -> spec.RawDexTrades:

        from ctc import config

        network = config.get_context_chain_id(context)
        vault = cls._pool_factories[network][0]
        if start_block is None:
            start_block = await evm.async_get_contract_creation_block(
                pool,
                context=context,
            )

        trades = await evm.async_get_events(
            vault,
            event_abi=balancer_utils.vault_event_abis['Swap'],
            start_block=start_block,
            end_block=end_block,
            start_time=start_time,
            end_time=end_time,
            verbose=verbose,
        )

        assets = await cls.async_get_pool_assets(pool)

        # filter by pool
        pool_id = await balancer_utils.async_get_pool_id(
            pool_address=pool, context=context
        )
        mask = trades['arg__poolId'] == pool_id
        trades = trades[mask]

        import polars as pl

        output: spec.RawDexTrades = {
            'block_number': trades['block_number'],
            'transaction_hash': trades['transaction_hash'],
            'recipient': None,
            'sold_id': pl.Series(
                [
                    assets.index(address)
                    for address in trades['arg__tokenIn'].to_list()
                ],
            ),
            'bought_id': pl.Series(
                [
                    assets.index(address)
                    for address in trades['arg__tokenOut'].to_list()
                ],
            ),
            'sold_amount': trades['arg__amountIn'].apply(int),
            'bought_amount': trades['arg__amountOut'].apply(int),
        }

        if include_timestamps:
            output['timestamp'] = await evm.async_get_block_timestamps(
                blocks=trades['block_number'].to_list(),
                context=context,
            )

        return output

    @classmethod
    async def async_get_pool_balance(
        cls,
        pool: spec.Address,
        asset: spec.Address,
        *,
        factory: spec.Address | None = None,
        normalize: bool = True,
        block: spec.BlockNumberReference | None = None,
        context: spec.Context = None,
    ) -> int | float:

        from ctc.protocols import balancer_utils

        pool_balances = await balancer_utils.async_get_pool_balances(
            pool_address=pool,
            block=block,
            normalize=normalize,
            context=context,
        )
        return pool_balances[asset]




from __future__ import annotations

import typing

from ctc import spec


def compute_ohlc(
    values: typing.Sequence[spec.Number],
    indices: typing.Sequence[spec.Number],
    *,
    bin_size: spec.Number | None = None,
    bins: typing.Sequence[spec.Number] | None = None,
    volumes: typing.Sequence[spec.Number] | None = None,
) -> spec.DataFrame:
    # TODO: add full support for non-integer bins

    import numpy as np
    import polars as pl

    # assemble data
    df_raw = pl.DataFrame({'index': indices, 'value': values})
    if volumes is not None:
        df_raw = df_raw.with_columns(pl.Series('volume', volumes))

    # set bins
    if bins is None:
        if bin_size is None:
            raise Exception('must specify bins or bin_size')
        bins_float = np.floor(df_raw['index'].to_numpy() / bin_size) * bin_size
        bins = bins_float.astype(int)
    else:
        bin_size = bins[1] - bins[0]  # type: ignore
    if bin_size is None:
        raise Exception('bin_size could not be determined')
    df_raw = df_raw.with_columns(pl.Series('bin', bins))

    # compute metrics
    columns = [
        pl.col('value').first().alias('open'),
        pl.col('value').max().alias('high'),
        pl.col('value').min().alias('low'),
        pl.col('value').last().alias('close'),
        pl.col('value').count().alias('count'),
    ]
    if volumes is not None:
        columns.append(pl.col('volume').sum().alias('volume'))
    df = df_raw.groupby('bin').agg(columns)

    # fill in missing bins
    all_bins = np.arange(
        bins[0], bins[-1] + bin_size, bin_size
    )
    n_to_add = len(all_bins) - len(df)
    if n_to_add > 0:
        last_close = df['close'][-1]
        row = [last_close, None, None, last_close, 0]
        if volumes is not None:
            row.append(0)
        missing_rows = pl.DataFrame(row * n_to_add, schema=df.schema)
        df = pl.concat([df, missing_rows])

    df = df.sort('bin')

    return df




from __future__ import annotations

import typing

import polars as pl

from ctc import spec


def concat(
    dfs: typing.Sequence[spec.DataFrame],
    **kwargs: typing.Any,
) -> spec.DataFrame:
    """like pl.concat, with support for objects"""

    # find object columns
    object_columns = []
    non_object_columns = []
    for column, datatype in dfs[0].schema.items():
        if datatype == pl.Object:
            object_columns.append(column)
        else:
            non_object_columns.append(column)

    # build object columns
    object_series = []
    for object_column in object_columns:
        values = []
        for df in dfs:
            values.extend(df[object_column].to_list())
        series = pl.Series(object_column, values, dtype=pl.Object)
        object_series.append(series)

    # concat non-object columns
    without_objects = [df[non_object_columns] for df in dfs]
    concated = pl.concat(without_objects, **kwargs)
    return concated.with_columns(object_series)




from __future__ import annotations

import asyncio
import decimal
import os
import typing

import polars as pl

import ctc.rpc
from ctc import spec
from ctc.toolbox import pl_utils

try:
    import pdp
    import tooljob.trackers.multifile_tracker
except ImportError:
    raise Exception('install pdp for dataset functionality')


class ExtractBlocksAndTransactions(pdp.BlockChunkJobs):
    extract_blocks: bool
    extract_transactions: bool
    tracker: tooljob.trackers.multifile_tracker.MultifileTracker

    def __init__(
        self,
        extract_blocks: bool = True,
        extract_transactions: bool = True,
        **kwargs: typing.Any,
    ) -> None:
        self.extract_blocks = extract_blocks
        self.extract_transactions = extract_transactions
        super().__init__(outputs=['blocks', 'transactions'], **kwargs)

    def execute_job(self, i: int) -> typing.Any:
        job_data = self.get_job_data(i)
        job_name = self.get_job_name(i)
        paths = self.tracker.get_job_output_paths(i)
        sync_extract_blocks_and_transactions(
            start_block=job_data['start_block'],
            end_block=job_data['end_block'],
            job_name=job_name,
            paths=paths,
            context=self.context,
            extract_blocks=self.extract_blocks,
            extract_transactions=self.extract_transactions,
        )


def sync_extract_blocks_and_transactions(
    *,
    start_block: int,
    end_block: int,
    job_name: str,
    paths: typing.Mapping[str, str],
    context: spec.Context,
    extract_blocks: bool,
    extract_transactions: bool,
) -> None:
    try:
        coroutine = async_extract_blocks_and_transactions(
            start_block=start_block,
            end_block=end_block,
            paths=paths,
            context=context,
            extract_blocks=extract_blocks,
            extract_transactions=extract_transactions,
        )
        asyncio.run(coroutine)
    except Exception as e:
        print('job', job_name, 'failed:' + str(e))
        raise e


async def async_extract_blocks_and_transactions(
    *,
    start_block: int,
    end_block: int,
    paths: typing.Mapping[str, str],
    context: spec.Context,
    extract_blocks: bool,
    extract_transactions: bool,
) -> None:

    # get paths
    if all(os.path.isfile(path) for path in paths.values()):
        return

    extract_blocks &= not os.path.isfile(paths['blocks'])
    extract_transactions &= not os.path.isfile(paths['transactions'])

    # collect data
    coroutines = []
    for block_number in range(start_block, end_block + 1):
        coroutine = ctc.rpc.async_eth_get_block_by_number(
            block_number=block_number,
            context=context,
            include_full_transactions=extract_transactions,
        )
        coroutines.append(coroutine)
    raw_blocks = await asyncio.gather(*coroutines)
    await ctc.rpc.async_close_http_session(context=context)

    # extract blocks
    if extract_blocks:
        blocks_df = raw_blocks_to_dataframe(raw_blocks)
        pl_utils.write_df(df=blocks_df, path=paths['blocks'])

    # extract transactions
    if extract_transactions:
        transactions_df = raw_transactions_to_dataframe(raw_blocks)
        pl_utils.write_df(df=transactions_df, path=paths['transactions'])


def raw_blocks_to_dataframe(
    raw_blocks: typing.Sequence[typing.Mapping[str, typing.Any]]
) -> spec.DataFrame:
    # process raw blocks
    blocks = []
    for raw_block in raw_blocks:
        block = {
            'base_fee_per_gas': raw_block['base_fee_per_gas'],
            'gas_limit': raw_block['gas_limit'],
            'gas_used': raw_block['gas_used'],
            'hash': raw_block['hash'],
            'miner': raw_block['miner'],
            'block_number': raw_block['number'],
            'timestamp': raw_block['timestamp'],
        }
        blocks.append(block)

    # create dataframe
    blocks_schema = {
        'base_fee_per_gas': pl.UInt64,
        'gas_limit': pl.UInt32,
        'gas_used': pl.UInt32,
        'hash': pl.Utf8,
        'miner': pl.Utf8,
        'block_number': pl.UInt32,
        'timestamp': pl.UInt32,
    }
    blocks_df = pl.DataFrame(blocks, schema=blocks_schema)

    # convert binary fields
    binary_fields = [
        'hash',
        'miner',
    ]
    blocks_df = blocks_df.with_columns(
        blocks_df[binary_fields].select(
            pl.col('*').str.slice(2).str.decode('hex')
        )
    )

    return blocks_df


def raw_transactions_to_dataframe(
    raw_blocks: typing.Sequence[typing.Mapping[str, typing.Any]]
) -> spec.DataFrame:
    """
    Not currently included:
    - signature data: v, r, s
    - receipt data: gas_used, gas_price, status
    - access_list
    """

    # process raw transactions
    transactions = []
    for raw_block in raw_blocks:
        for raw_tx in raw_block['transactions']:
            to = raw_tx.get('to')
            if to is None:
                to = '0x0000000000000000000000000000000000000000'
            transaction = {
                'transaction_hash': raw_tx['hash'],
                'block_number': raw_tx['block_number'],
                'transaction_index': raw_tx['transaction_index'],
                'to_address': to,
                'from_address': raw_tx['from'],
                'value_float': float(raw_tx['value']),
                'value_decimal': decimal.Decimal(raw_tx['value']),
                'input': raw_tx['input'],
                'nonce': raw_tx['nonce'],
                'transaction_type': raw_tx['type'],
                'gas_limit': raw_tx['gas'],
                'gas_priority_max': raw_tx.get('max_priority_fee_per_gas'),
                'gas_price_max': raw_tx.get('max_fee_per_gas'),
                'gas_price': raw_tx.get('gas_price'),
            }
            transactions.append(transaction)

    # create dataframe
    transactions_schema = {
        'transaction_hash': pl.Utf8,
        'block_number': pl.UInt32,
        'transaction_index': pl.UInt32,
        'to_address': pl.Utf8,
        'from_address': pl.Utf8,
        'value_float': pl.Float64,
        'value_decimal': pl.Decimal,
        'input': pl.Utf8,
        'nonce': pl.UInt32,
        'transaction_type': pl.UInt8,
        'gas_limit': pl.Int64,
        'gas_priority_max': pl.Int64,
        'gas_price_max': pl.Int64,
        'gas_price': pl.Int64,
    }
    transactions_df = pl.DataFrame(transactions, schema=transactions_schema)

    # convert binary fields
    binary_fields = [
        'transaction_hash',
        'to_address',
        'from_address',
        'input',
    ]
    transactions_df = transactions_df.with_columns(
        transactions_df[binary_fields].select(
            pl.col('*').str.slice(2).str.decode('hex')
        )
    )

    return transactions_df




from __future__ import annotations

import typing
from typing_extensions import TypedDict, Literal, NotRequired

from . import address_types
from . import binary_types
from . import transaction_types


#
# # block references
#

BlockHash = binary_types.PrefixHexData

BlockNumberName = typing.Union[
    Literal['latest'],
    Literal['earliest'],
    Literal['pending'],
]

# anything that can be converted to an int without node querying
RawBlockNumber = typing.Union[typing.SupportsRound, binary_types.HexData, str]


class RawBlock(TypedDict):
    pass


# an int or block number name
StandardBlockNumber = typing.Union[int, BlockNumberName]

# anything that refers to a block number, raw or standard
BlockNumberReference = typing.Union[RawBlockNumber, StandardBlockNumber]

# any reference to a block
BlockReference = typing.Union[BlockNumberReference, BlockHash]


class BlockRange(TypedDict):
    start_block: StandardBlockNumber
    end_block: StandardBlockNumber
    open_start: bool
    open_end: bool


class BlockSample(TypedDict):
    start_block: StandardBlockNumber
    end_block: StandardBlockNumber
    block_interval: typing.Union[int, None]
    open_start: bool
    open_end: bool


#
# # block data
#

# block returned from RPC request
class RPCBlock(TypedDict):
    base_fee_per_gas: NotRequired[int | None]
    difficulty: int
    extra_data: binary_types.PrefixHexData
    gas_limit: int
    gas_used: int
    hash: BlockHash
    logs_bloom: binary_types.PrefixHexData
    miner: address_types.Address
    mix_hash: BlockHash
    nonce: binary_types.PrefixHexData
    number: int
    parent_hash: BlockHash
    receipts_root: binary_types.PrefixHexData
    sha3_uncles: binary_types.PrefixHexData
    size: int
    state_root: binary_types.PrefixHexData
    timestamp: int
    total_difficulty: str
    transactions: typing.Union[
        typing.List[transaction_types.TransactionHash],
        typing.List[transaction_types.RPCTransaction],
    ]
    transactions_root: binary_types.PrefixHexData
    uncles: typing.List[BlockHash]


class DBBlock(TypedDict):
    number: int
    hash: str
    timestamp: int
    miner: str
    extra_data: str
    base_fee_per_gas: int | None
    gas_limit: int
    gas_used: int




from __future__ import annotations

import toolcli

from ctc.protocols import yearn_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_yearn_vaults_command,
        'help': 'display summary of Yearn vaults',
        'args': [
            {
                'name': 'query',
                'help': 'address or name of vault to show details of',
                'nargs': '?',
            },
            {
                'name': '--sort',
                'help': 'vault field to sort by',
            },
            {
                'name': '-n',
                'help': 'number of vaults to display',
                'type': int,
                'default': 20,
            },
            {
                'name': '--min-apy',
                'help': 'filter vaults by minimum APY',
                'type': float,
            },
            {
                'name': '--min-apr',
                'help': 'filter vaults by minimum APR',
                'type': float,
            },
            {
                'name': '--min-tvl',
                'help': 'filter vaults by minimum TVL',
                'type': float,
            },
            {
                'name': ['-v', '--verbose'],
                'help': 'output additional information',
                'action': 'store_true',
            },
            {
                'name': '--harvests',
                'help': 'show harvests when viewing individual vault',
                'action': 'store_true',
            },
        ],
        'examples': [
            '',
            '--min-tvl 1000000',
            '--sort TVL',
            '0x9d409a0a012cfba9b15f6d4b36ac57a46966ab9a',
            '0x9d409a0a012cfba9b15f6d4b36ac57a46966ab9a -v',
            '0x9d409a0a012cfba9b15f6d4b36ac57a46966ab9a --harvests',
        ],
    }


async def async_yearn_vaults_command(
    *,
    query: str | None,
    sort: str | None,
    n: int | None,
    min_apy: float | None,
    min_apr: float | None,
    min_tvl: float | None,
    verbose: bool,
    harvests: bool,
) -> None:

    if query is not None:
        await yearn_utils.async_print_vault_summary(
            query=query,
            verbose=verbose,
            network='ethereum',
            show_harvests=harvests,
        )

    else:
        await yearn_utils.async_print_vaults_summary(
            sort_by=sort,
            min_tvl=min_tvl,
            min_apy=min_apy,
            min_apr=min_apr,
            n=n,
            verbose=verbose,
            network='ethereum',
        )



from __future__ import annotations

import toolcli
import toolstr

from ctc import cli
from .. import safe_summary


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_gnosis_command,
        'help': 'print information about a gnosis safe',
        'args': [
            {'name': 'address', 'help': 'address of gnosis safe'},
            {
                'name': ['--verbose', '-v'],
                'help': 'display additional information',
                'action': 'store_true',
            },
        ],
        'examples': [
            '0x245cc372c84b3645bf0ffe6538620b04a217988b',
            '0x245cc372c84b3645bf0ffe6538620b04a217988b --verbose',
        ],
    }


async def async_gnosis_command(address: str, verbose: bool) -> None:
    await safe_summary.async_print_safe_summary(address, verbose=verbose)
    if not verbose:
        styles = cli.get_cli_styles()
        toolstr.print(
            'use --verbose to display detailed information',
            style=styles['comment'],
        )



from __future__ import annotations

from ctc import spec


async def async_get_irm_blocks_per_year(
    interest_rate_model: spec.Address,
    block: spec.BlockNumberReference = 'latest',
) -> int:

    return 2102400
    # try:
    #     return await rpc.async_eth_call(
    #         to_address=interest_rate_model,
    #         block_number=block,
    #         function_abi=rari_abis.interest_rate_model_abis['blocksPerYear'],
    #     )
    # except KeyError:
    #     return await rpc.async_eth_call(
    #         to_address=rari_abis.contract_examples['InterestRateModel'],
    #         block_number=block,
    #         function_abi=rari_abis.interest_rate_model_abis['blocksPerYear'],
    #     )



from __future__ import annotations

import toolcli

from .. import llama_yields


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_llama_pool_command,
        'help': 'output data about a pool tracked by Defi Llama',
        'args': [
            {
                'name': 'pool',
                'help': 'id of pool (obtain using --show-id on llama pools cmd)',
            },
        ],
        'examples': [
            '747c1d2a-c668-4682-b9f9-296708a3dd90',
        ],
    }


async def async_llama_pool_command(pool: str) -> None:
    await llama_yields.async_print_pool_yield_summary(pool)




import toolsql

from ctc import db
from ctc.protocols import fourbyte_utils

import conftest


example_data = [
    {
        'id': 145,
        'created_at': '2016-07-09T03:58:28.234977Z',
        'text_signature': 'transfer(address,uint256)',
        'hex_signature': '0xa9059cbb',
    },
    {
        'id': 179,
        'created_at': '2016-07-09T03:58:45.230129Z',
        'hex_signature': '0x18160ddd',
        'text_signature': 'totalSupply()',
    },
    {
        'id': 31780,
        'created_at': '2018-05-11T08:39:29.708250Z',
        'text_signature': 'many_msg_babbage(bytes1)',
        'hex_signature': '0xa9059cbb',
    },
    {
        'id': 161159,
        'created_at': '2019-03-22T19:13:17.314877Z',
        'text_signature': 'transfer(bytes4[9],bytes5[6],int48[11])',
        'hex_signature': '0xa9059cbb',
    },
    {
        'id': 313067,
        'created_at': '2021-10-20T05:29:13.555535Z',
        'text_signature': 'func_2093253501(bytes)',
        'hex_signature': '0xa9059cbb',
    },
]


async def test_fourbyte_crud():

    db_config = conftest.get_test_db_config()
    db_schema = db.get_raw_schema(schema_name='4byte')
    toolsql.create_db(
        db_config=db_config,
        db_schema=db_schema,
        if_not_exists=True,
        confirm=True,
    )

    # insert data
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await fourbyte_utils.async_upsert_function_signature(
                conn=conn,
                function_signature=datum,
            )

    # get data individually
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            db_data = await fourbyte_utils.async_select_function_signatures(
                conn=conn,
                id=datum['id'],
            )
            db_datum = db_data[0]
            for key, target_value in datum.items():
                assert target_value == db_datum[key]

    # get data collectively
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            db_datas = (
                await fourbyte_utils.async_select_function_signatures(
                    conn=conn,
                    hex_signature=datum['hex_signature'],
                )
            )
            assert len(db_datas) > 0

    # delete entries one by one
    async with toolsql.async_connect(db_config) as conn:
        for datum in example_data:
            await fourbyte_utils.async_delete_function_signatures(
                conn=conn,
                text_signature=datum['text_signature'],
            )

    # ensure all entries deleted
    async with toolsql.async_connect(db_config) as conn:
        db_datas = await fourbyte_utils.async_select_function_signatures(
            conn=conn,
        )
        assert len(db_datas) == 0



import time

import pytest

from ctc import rpc


@pytest.mark.asyncio
async def test_eth_new_filter():
    new_filter = await rpc.async_eth_new_filter()
    assert len(new_filter) > 0 and new_filter.startswith('0x')


@pytest.mark.asyncio
async def test_eth_new_block_filter():
    new_filter = await rpc.async_eth_new_block_filter()
    assert len(new_filter) > 0 and new_filter.startswith('0x')
    # blocks = await rpc.async_eth_get_filter_changes(new_filter)


@pytest.mark.asyncio
async def test_eth_new_pending_transaction_filter():
    new_filter = await rpc.async_eth_new_pending_transaction_filter()
    assert len(new_filter) > 0 and new_filter.startswith('0x')
    # pending_transactions = await rpc.async_eth_get_filter_changes(new_filter)


@pytest.mark.asyncio
async def test_eth_uninstall_filter():
    new_filter = await rpc.async_eth_new_filter()
    await rpc.async_eth_uninstall_filter(new_filter)


# @pytest.mark.asyncio
# async def test_eth_get_filter_changes():
#     new_filter = await rpc.async_eth_new_pending_transaction_filter()
#     time.sleep(3)
#     pending_transactions = await rpc.async_eth_get_filter_changes(new_filter)
#     assert len(pending_transactions) > 0


# @pytest.mark.asyncio
# async def test_eth_get_filter_logs():
#     new_filter = await rpc.async_eth_new_filter(
#         topics=None,
#         start_block=None,
#         end_block=None,
#     )
#     new_filter = await rpc.async_eth_new_filter(
#         topics=[
#             '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'
#         ],
#         start_block=12345677,
#         end_block=12345679,
#     )
#     logs = await rpc.async_eth_get_filter_logs(new_filter)
#     assert len(logs) == 375


@pytest.mark.asyncio
async def test_eth_get_logs():
    logs = await rpc.async_eth_get_logs(
        topics=[
            '0xddf252ad1be2c89b69c2b068fc378daa952ba7f163c4a11628f55a4df523b3ef'
        ],
        start_block=12345677,
        end_block=12345679,
    )
    assert len(logs) == 375



from __future__ import annotations

import asyncio
import typing

from ctc.protocols import chainlink_utils
from ctc import rpc
from ctc import spec

from . import aave_spec


async def async_get_asset_price(
    asset: spec.Address,
    *,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    units: typing.Literal['usd', 'eth'] = 'usd',
    context: spec.Context = None,
) -> int | float:

    oracle = aave_spec.get_aave_address('PriceOracle', context=context)
    price_coroutine = rpc.async_eth_call(
        to_address=oracle,
        block_number=block,
        function_name='getAssetPrice',
        function_parameters=[asset],
        context=context,
    )

    if units == 'eth':
        price: int | float = await price_coroutine

    elif units == 'usd':
        eth_usd_coroutine = chainlink_utils.async_get_eth_price(
            block=block,
            normalize=True,
            context=context,
        )
        asset_price, eth_usd = await asyncio.gather(
            price_coroutine, eth_usd_coroutine
        )
        price = asset_price * eth_usd

    else:
        raise Exception('unknown units: ' + str(units))

    if not isinstance(price, (int, float)):
        raise Exception('invalid rpc result')

    if normalize:
        price = price / 1e18

    return price


async def async_get_asset_prices(
    assets: typing.Sequence[spec.Address],
    *,
    block: spec.BlockNumberReference = 'latest',
    normalize: bool = True,
    units: typing.Literal['usd', 'eth'] = 'usd',
    context: spec.Context = None,
) -> typing.Sequence[int | float]:

    oracle = aave_spec.get_aave_address('PriceOracle', context=context)
    price_coroutine = rpc.async_batch_eth_call(
        to_address=oracle,
        block_number=block,
        function_name='getAssetPrice',
        function_parameter_list=[[asset] for asset in assets],
        context=context,
    )

    if units == 'eth':
        prices = await price_coroutine

    elif units == 'usd':
        eth_usd_coroutine = chainlink_utils.async_get_eth_price(
            block=block, normalize=True, context=context,
        )
        asset_prices, eth_usd = await asyncio.gather(
            price_coroutine, eth_usd_coroutine
        )
        prices = [asset_price * eth_usd for asset_price in asset_prices]

    else:
        raise Exception('unknown units: ' + str(units))

    if normalize:
        prices = [price / 1e18 for price in prices]

    return prices


async def async_get_asset_price_by_block(
    asset: spec.Address,
    *,
    blocks: typing.Sequence[spec.BlockNumberReference],
    normalize: bool = True,
    units: typing.Literal['usd', 'eth'] = 'usd',
    context: spec.Context = None,
) -> typing.Sequence[int | float]:
    coroutines = [
        async_get_asset_price(
            asset=asset,
            block=block,
            normalize=normalize,
            units=units,
            context=context,
        )
        for block in blocks
    ]
    return await asyncio.gather(*coroutines)




from __future__ import annotations

import toolcli

from ctc import evm
from ctc import spec
from ctc.protocols import ens_utils


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': async_owner_command,
        'help': 'output owner of ENS name',
        'args': [
            {'name': 'name', 'help': 'ENS name'},
            {'name': '--block', 'help': 'block number'},
        ],
        'examples': [
            'vitalik.eth',
            'vitalik.eth --block 14000000',
        ],
    }


async def async_owner_command(
    name: str, block: spec.BlockNumberReference
) -> None:
    if block is not None:
        block = evm.standardize_block_number(block)
    owner = await ens_utils.async_get_owner(name, block=block)
    print(owner)



from __future__ import annotations

import asyncio
import typing

import toolsql

from ctc import spec
from ... import schema_utils


async def async_upsert_erc20_metadata(
    *,
    context: spec.Context,
    address: spec.Address,
    symbol: str | None = None,
    decimals: int | None = None,
    name: str | None = None,
    upsert: bool = True,
    conn: toolsql.AsyncConnection,
) -> None:

    # construct row
    row = {
        'address': address.lower(),
        'symbol': symbol,
        'decimals': decimals,
        'name': name,
    }
    row = {k: v for k, v in row.items() if v is not None}

    # get table name
    table = schema_utils.get_table_schema('erc20_metadata', context=context)

    # insert data
    await toolsql.async_insert(
        conn=conn,
        table=table,
        row=row,
        upsert=upsert,
    )


async def async_upsert_erc20s_metadata(
    *,
    erc20s_metadata: typing.Sequence[spec.ERC20Metadata],
    context: spec.Context,
    conn: toolsql.AsyncConnection,
) -> None:
    coroutines = [
        async_upsert_erc20_metadata(conn=conn, context=context, **metadata)
        for metadata in erc20s_metadata
    ]
    await asyncio.gather(*coroutines)


async def async_select_erc20_metadata(
    address: spec.Address | None = None,
    *,
    symbol: str | None = None,
    case_insensitive_symbol: bool = False,
    context: spec.Context | None = None,
    conn: toolsql.AsyncConnection,
) -> spec.ERC20Metadata | None:

    table = schema_utils.get_table_schema('erc20_metadata', context=context)

    if address is not None:
        query: typing.Mapping[str, typing.Any] = {
            'where_equals': {'address': address.lower()}
        }
    elif symbol is not None:
        if case_insensitive_symbol:
            query = {'where_ilike': {'symbol': symbol}}
        else:
            query = {'where_equals': {'symbol': symbol}}
    else:
        raise Exception('must specify address or symbol')

    erc20_metadata: spec.ERC20Metadata = await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        output_format='single_dict_or_none',
        **query,
    )

    return erc20_metadata


async def async_select_erc20s_metadata(
    addresses: typing.Sequence[spec.Address],
    *,
    context: spec.Context | None = None,
    conn: toolsql.AsyncConnection,
) -> typing.Sequence[spec.ERC20Metadata | None] | None:

    table = schema_utils.get_table_schema('erc20_metadata', context=context)
    results: typing.Sequence[spec.ERC20Metadata] = await toolsql.async_select(  # type: ignore
        conn=conn,
        table=table,
        where_in={'address': [address.lower() for address in addresses]},
    )

    if results is None:
        return None

    # package into output
    results_by_address = {result['address']: result for result in results}

    return [results_by_address.get(address) for address in addresses]


async def async_delete_erc20_metadata(
    address: spec.Address,
    *,
    context: spec.Context,
    conn: toolsql.AsyncConnection,
) -> None:

    table = schema_utils.get_table_schema('erc20_metadata', context=context)
    await toolsql.async_delete(
        table=table,
        conn=conn,
        where_equals={'address': address.lower()},
    )


async def async_delete_erc20s_metadata(
    addresses: typing.Sequence[spec.Address],
    *,
    context: spec.Context,
    conn: toolsql.AsyncConnection,
) -> None:

    table = schema_utils.get_table_schema('erc20_metadata', context=context)
    await toolsql.async_delete(
        table=table,
        conn=conn,
        where_in={'address': [address.lower() for address in addresses]},
    )




"""
https://eips.ethereum.org/EIPS/eip-712
https://eips.ethereum.org/EIPS/eip-191

TODO:
- handle nested struct types
"""

from __future__ import annotations

import typing

from ctc import spec
from ... import abi_utils
from .. import format_utils
from .. import hash_utils
from . import signature_creation
from . import signature_recovery


def hash_eip712_struct(
    struct_data: typing.Mapping[str, typing.Any],
    *,
    struct_type: spec.Eip712StructType,
    domain: typing.Mapping[str, typing.Any],
    output_format: spec.BinaryFormat = 'prefix_hex',
) -> spec.Data:
    """encode structured data message for use with EIP-712"""

    as_bytes = (
        '\x19\x01'.encode()
        + get_domain_separator(domain, 'binary')
        + _hash_struct(struct_data=struct_data, struct_type=struct_type)
    )
    return hash_utils.keccak(as_bytes, output_format)


def sign_eip712_struct(
    struct_data: typing.Mapping[str, typing.Any],
    *,
    private_key: spec.Data,
    struct_type: spec.Eip712StructType,
    domain: typing.Mapping[str, typing.Any],
) -> tuple[int, int, int]:
    """sign EIP-712 struc data"""

    message_hash = hash_eip712_struct(
        struct_data=struct_data,
        struct_type=struct_type,
        domain=domain,
    )

    if domain.get('chain_id') is not None:
        chain_id = domain.get('chain_id')
    elif domain.get('chainId') is not None:
        chain_id = domain.get('chainId')
    else:
        chain_id = None

    return signature_creation.sign_message_hash(
        message_hash=message_hash,
        private_key=private_key,
        chain_id=chain_id,
    )


def verify_eip712_signature(
    signature: spec.Signature,
    *,
    struct_data: typing.Mapping[str, typing.Any],
    struct_type: spec.Eip712StructType,
    domain: typing.Mapping[str, typing.Any],
    public_key: spec.Data | None = None,
    address: spec.Data | None = None,
) -> bool:
    """verify signed EIP-712 data"""

    message_hash = hash_eip712_struct(
        struct_data=struct_data,
        struct_type=struct_type,
        domain=domain,
    )

    return signature_recovery.verify_signature(
        signature=signature,
        message_hash=message_hash,
        public_key=public_key,
        address=address,
    )


def _hash_struct(
    struct_data: typing.Mapping[str, typing.Any],
    struct_type: spec.Eip712StructType,
) -> bytes:
    """compute struct hash for use with EIP-712"""

    return hash_utils.keccak(
        _hash_struct_type(struct_type)
        + _encode_struct_data(struct_data=struct_data, struct_type=struct_type),
        'binary',
    )


def _hash_struct_type(struct_type: spec.Eip712StructType) -> bytes:
    """compute struct type hash for use with EIP-712"""

    struct_fields_str = ','.join(
        type + ' ' + name for name, type in struct_type['fields'].items()
    )
    struct_type_str = struct_type['name'] + '(' + struct_fields_str + ')'
    return hash_utils.keccak_text(struct_type_str, 'binary')


def _encode_struct_data(
    struct_data: typing.Mapping[str, typing.Any],
    struct_type: spec.Eip712StructType,
) -> bytes:
    """encode struct data for use with EIP-712"""

    all_encoded = bytes()
    for name, type in struct_type['fields'].items():
        item = _encode_datum(value=struct_data[name], type=type)
        all_encoded += item
    return all_encoded


def _encode_datum(value: typing.Any, type: spec.ABIDatatypeStr) -> bytes:
    """encode scalar datum for use with EIP-712"""

    if '[' in type:
        # array
        scalar_type = type[: type.index('[')]
        encoded = bytes()
        for subvalue in value:
            encoded += _encode_datum(subvalue, scalar_type)
        return hash_utils.keccak(encoded, 'binary')
    elif type == 'bool':
        return abi_utils.abi_encode(int(value), 'uint256')
    elif type == 'address':
        converted = format_utils.binary_convert(value, 'integer')
        return abi_utils.abi_encode(converted, 'uint160')
    elif type.startswith('int'):
        return abi_utils.abi_encode(value, 'int256')
    elif type.startswith('uint'):
        return abi_utils.abi_encode(value, 'uint256')
    elif type.startswith('bytes') and type[5:].isdecimal():
        return typing.cast(bytes, value) + b'\x00' * (32 - len(value))
    elif type == 'bytes':
        return hash_utils.keccak(value, 'binary')
    elif type == 'string':
        return hash_utils.keccak_text(value, 'binary')
    elif type == 'struct':
        raise NotImplementedError('nested struct encoding')
    else:
        raise Exception('type encoding unknown')


@typing.overload
def get_domain_separator(
    domain: typing.Mapping[str, typing.Any],
    output_format: typing.Literal['binary'],
    *,
    custom_fields: typing.Mapping[str, spec.ABIDatatypeStr] | None = None,
) -> bytes:
    ...


@typing.overload
def get_domain_separator(
    domain: typing.Mapping[str, typing.Any],
    output_format: typing.Literal['prefix_hex', 'raw_hex'],
    *,
    custom_fields: typing.Mapping[str, spec.ABIDatatypeStr] | None = None,
) -> str:
    ...


@typing.overload
def get_domain_separator(
    domain: typing.Mapping[str, typing.Any],
    output_format: typing.Literal['integer'] = 'integer',
    *,
    custom_fields: typing.Mapping[str, spec.ABIDatatypeStr] | None = None,
) -> int:
    ...


def get_domain_separator(
    domain: typing.Mapping[str, typing.Any],
    output_format: spec.BinaryFormat = 'prefix_hex',
    *,
    custom_fields: typing.Mapping[str, spec.ABIDatatypeStr] | None = None,
) -> spec.Data:
    """compute domain separator for use with EIP-712"""

    # template of standard EIP-712 fields
    domain_separator_fields: typing.Mapping[str, spec.ABIDatatypeStr] = {
        'name': 'string',
        'version': 'string',
        'chainId': 'uint256',
        'verifyingContract': 'address',
        'salt': 'bytes32',
    }

    # convert snake case to camel case
    if 'chain_id' in domain and 'chainId' in domain:
        raise Exception('should only specify one of chain_id or chainId')
    if 'verifying_contract' in domain and 'verifyingContract' in domain:
        raise Exception(
            'should only specify one of verifying_contract or verifyingContract'
        )
    if 'chain_id' in domain:
        domain = {
            (k if k != 'chain_id' else 'chainId'): v for k, v in domain.items()
        }
    if 'verifying_contract' in domain:
        domain = {
            (k if k != 'verifying_contract' else 'verifyingContract'): v
            for k, v in domain.items()
        }

    # add in custom fields if present
    if custom_fields is not None:
        domain_separator_fields = dict(domain_separator_fields, **custom_fields)

    # adjust type for whichever domain fields are present
    domain_separator_fields = {
        k: v for k, v in domain_separator_fields.items() if k in domain
    }

    # create domain separator struct type
    domain_separator_type: spec.Eip712StructType = {
        'name': 'EIP712Domain',
        'fields': domain_separator_fields,
    }

    # ensure that no extra fields are present in domain
    for key in domain.keys():
        if key not in domain_separator_fields:
            raise Exception('unrecognized domain_separator field: ' + str(key))

    # compute struct hash of domain
    as_bytes = _hash_struct(
        struct_data=domain,
        struct_type=domain_separator_type,
    )
    return format_utils.binary_convert(as_bytes, output_format)



from __future__ import annotations

import typing

from ctc import rpc
from ctc import spec
from . import safe_spec


async def async_get_safe_owners(
    address: spec.Address,
    *,
    block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> typing.Sequence[spec.Address]:
    result = await rpc.async_eth_call(
        to_address=address,
        function_abi=safe_spec.function_abis['getOwners'],
        block_number=block,
        context=context,
    )
    if not isinstance(result, (tuple, list)) or not all(
        isinstance(item, str) for item in result
    ):
        raise Exception('invalid rpc result')
    return result


async def async_get_safe_threshold(
    address: spec.Address,
    *,
    block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> int:
    result = await rpc.async_eth_call(
        to_address=address,
        function_abi=safe_spec.function_abis['getThreshold'],
        block_number=block,
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result


async def async_get_safe_nonce(
    address: spec.Address,
    *,
    block: spec.BlockReference | None = None,
    context: spec.Context = None,
) -> int:
    result = await rpc.async_eth_call(
        to_address=address,
        function_abi=safe_spec.function_abis['nonce'],
        block_number=block,
        context=context,
    )
    if not isinstance(result, int):
        raise Exception('invalid rpc result')
    return result



from __future__ import annotations

from ctc import spec
from . import erc721_spec


async def async_get_erc721_total_supply(
    token: spec.Address,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> int:
    """get total supply of erc721 token"""
    from ctc import rpc

    result = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc721_spec.erc721_function_abis['totalSupply'],
        block_number=block,
        context=context,
    )

    if not isinstance(result, int):
        raise Exception('invalid return type')

    return result


async def async_get_erc721_owner(
    token: spec.Address,
    token_id: int,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> spec.Address:
    """get owner of erc721 by token_id"""
    from ctc import rpc

    result = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc721_spec.erc721_function_abis['ownerOf'],
        function_parameters=[token_id],
        block_number=block,
        context=context,
    )

    if not isinstance(result, str):
        raise Exception('invalid return type')

    return result


async def async_get_erc721_balance(
    token: spec.Address,
    wallet: spec.Address,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> int:
    """get number of erc721 tokens owner by wallet"""
    from ctc import rpc

    result = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc721_spec.erc721_function_abis['balanceOf'],
        function_parameters=[wallet],
        block_number=block,
        context=context,
    )

    if not isinstance(result, int):
        raise Exception('invalid return type')

    return result


async def async_get_erc721_approved(
    token: spec.Address,
    token_id: int,
    *,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> spec.Address:
    """return address approved for erc721 token_id"""
    from ctc import rpc

    result = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc721_spec.erc721_function_abis['getApproved'],
        function_parameters=[token_id],
        block_number=block,
        context=context,
    )

    if not isinstance(result, str):
        raise Exception('invalid return type')

    return result


async def async_get_erc721_approved_for_all(
    token: spec.Address,
    *,
    owner: spec.Address,
    operator: spec.Address,
    block: spec.BlockNumberReference | None = None,
    context: spec.Context = None,
) -> bool:
    """return whether owner has approved operator for all erc721 transfers"""
    from ctc import rpc

    result = await rpc.async_eth_call(
        to_address=token,
        function_abi=erc721_spec.erc721_function_abis['isApprovedForAll'],
        function_parameters=[owner, operator],
        block_number=block,
        context=context,
    )

    if not isinstance(result, bool):
        raise Exception('invalid return type')

    return result



from __future__ import annotations

import toolcli
import toolsql

from ctc import config
from ctc import db


def get_command_spec() -> toolcli.CommandSpec:
    return {
        'f': drop_command,
        'args': [
            {'name': 'schema_name', 'help': 'schema to drop'},
            {
                'name': 'network',
                'nargs': '?',
                'help': 'drop schema for single network, leave blank for all networks',
            },
            {
                'name': '--confirm',
                'help': 'confirm that schema should be dropped',
                'action': 'store_true',
            },
        ],
        'help': 'drop schema from db',
        'examples': [
            'blocks mainnnet',
            'blocks arbitrum',
        ],
        'hidden': True,
    }


def drop_command(
    *, schema_name: str, network: str | int, confirm: bool
) -> None:

    if isinstance(network, str):
        if network.isdigit() or (network[0] == '-' and network[1:].isdigit()):
            network = int(network)

    context = config.create_user_input_context(network=network)

    db_config = config.get_context_db_config(
        context=context, schema_name=schema_name
    )
    with toolsql.connect(db_config) as conn:
        db.drop_schema(
            schema_name=schema_name,
            confirm=confirm,
            context=context,
            conn=conn,
        )




from __future__ import annotations

import typing

import toolsql

from ctc import config
from ctc import spec
from . import dex_pools_statements


async def async_intake_dex_pools(
    *,
    factory: spec.Address,
    dex_pools: typing.Sequence[spec.DexPool],
    last_scanned_block: int | None = None,
    context: spec.Context = None,
) -> None:
    """intake dex pools into database

    - this function should always be provided with the complete set of dex pools
        that have been created since the previous last_scanned_block
    - all dex pools should have the same factory
    """

    # verify factories are identical
    for dex_pool in dex_pools:
        if dex_pool['factory'].lower() != factory:
            raise Exception('all dex pools should be produced by same factory')

    # obtain last scanned block if not provided
    if last_scanned_block is None:
        if len(dex_pools) == 0:
            return
        last_scanned_block = max(
            dex_pool['creation_block'] for dex_pool in dex_pools
        )

    db_config = config.get_context_db_config(
        schema_name='dex_pools',
        context=context,
    )
    async with toolsql.async_connect(db_config) as conn:
        await dex_pools_statements.async_upsert_dex_pools(
            dex_pools=dex_pools,
            conn=conn,
            context=context,
        )
        await dex_pools_statements.async_upsert_dex_pool_factory_query(
            factory=factory,
            last_scanned_block=last_scanned_block,
            conn=conn,
            context=context,
        )
